import json

from django.conf import settings
from django.shortcuts import reverse

from topobank.supplib.versions import get_versions

from .utils import current_selection_as_basket_items

HOME_URL = reverse('home')
SELECT_URL = reverse('ce_ui:select')
UNSELECT_ALL_URL = reverse('ce_ui:unselect-all')


def fixed_tabs_processor(request):
    """Adds fixed tabs.

    Parameters
    ----------
    request

    Returns
    -------
    Dict with extra context, having a key 'fixed_tabs' containing a list of tabs.

    Each tab is a dict with the following form:

    {
        'login_required': True,  # boolean, if True, tab is only shown to logged-in users
        'title': 'Tab title shown on the tab',
        'icon': 'edit',  # a fontawesome icon name,
        'icon_style_prefix': 'fas',  # class used as prefix for icons, default: 'fas' (=solid)
        'href': '',  # A URL pointing to the view for the tab
        'active': False,  #  a boolean; True means the tab is active
    }
    """

    tabs = []
    if settings.TABNAV_DISPLAY_HOME_TAB:
        tabs += [{
            'login_required': False,
            'title': '',  # no text
            'icon': 'home',
            'href': HOME_URL,
            'active': request.path == HOME_URL,
            'tooltip': "Welcome to contact.engineering",
            'show_basket': False,
        }]

    # This is the datasets tab
    tabs += [{
        'login_required': False,
        'title': 'Datasets',
        'icon': 'check-square',
        'icon_style_prefix': 'far',
        'href': SELECT_URL,
        'active': request.path == SELECT_URL,
        'tooltip': "Select surfaces and topographies for analysis or create new surfaces",
        'show_basket': True,
    }]

    # Add default value for icon_style_prefix if missing
    for tab in tabs:
        tab.setdefault('icon_style_prefix', 'fas')

    return dict(fixed_tabs=tabs)


def versions_processor(request):
    return dict(versions=get_versions())


def basket_processor(request):
    """Return JSON with select surfaces and topographies.

    Parameters
    ----------
    request

    Returns
    -------
    Dict with extra context, a key 'basket_items_json'
    which encodes all selected topographies and surfaces such they can be
    displayed on top of each page. See also base.html.
    """
    basket_items = current_selection_as_basket_items(request)

    return dict(basket_items_json=json.dumps(basket_items),
                num_basket_items=len(basket_items),
                unselect_all_url=UNSELECT_ALL_URL)



import pytest
from django.shortcuts import reverse
from django.utils import timezone

from topobank.utils import assert_in_content, assert_not_in_content
from termsandconditions.models import TermsAndConditions


#
# The code in these tests rely on a middleware which replaces
# Django's AnonymousUser by the one of django guardian
#


@pytest.mark.django_db
def test_anonymous_user_only_published_as_default(client):
    response = client.get(reverse('ce_ui:select'))
    assert_not_in_content(response, 'All accessible datasets')
    assert_not_in_content(response, 'Only own datasets')
    assert_not_in_content(response, 'Only datasets shared with you')
    assert_not_in_content(response, 'Only datasets shared by you')
    assert_not_in_content(response, 'Only datasets published by you')
    assert_in_content(response, 'Only datasets published by others')


@pytest.mark.django_db
def test_terms_conditions_as_anonymous(client, handle_usage_statistics):
    # Install terms and conditions for test
    TermsAndConditions.objects.create(slug='test-terms', name="Test of T&amp;C",
                                      text="some text", date_active=timezone.now())

    response = client.get(reverse('terms'))
    assert_in_content(response, "Test of T&amp;C")

    response = client.get(reverse('tc_accept_specific_page', kwargs=dict(slug='test-terms')))
    assert_in_content(response, "some text")



import logging
from html import unescape
from io import BytesIO

from django.core.exceptions import PermissionDenied
from django.db.models import Q, Value, TextField
from django.db.models.functions import Replace
from django.http import HttpResponse
from django.urls import reverse
from django.views.generic import DetailView, TemplateView

from rest_framework import generics
from rest_framework.decorators import api_view, permission_classes
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.utils.urls import remove_query_param, replace_query_param

from guardian.compat import get_user_model as guardian_user_model
from guardian.shortcuts import get_objects_for_user
from guardian.utils import get_user_obj_perms_model

from termsandconditions.models import TermsAndConditions
from termsandconditions.views import TermsView as OrigTermsView, AcceptTermsView
from trackstats.models import Metric, Period

from topobank.analysis.models import AnalysisFunction
from topobank.analysis.registry import AnalysisRegistry
from topobank.manager.containers import write_surface_container
from topobank.manager.models import Topography, Surface, TagModel
from topobank.manager.utils import surfaces_for_user, subjects_from_base64
from topobank.manager.utils import get_reader_infos
from topobank.usage_stats.utils import increase_statistics_by_date, increase_statistics_by_date_and_object, \
    current_statistics
from topobank.users.models import User

from .serializers import TagSearchSerizalizer, SurfaceSearchSerializer
from .utils import instances_to_selection, selected_instances, tags_for_user, current_selection_as_basket_items, \
    filtered_topographies, get_search_term, get_category, get_sharing_status, get_tree_mode, \
    filter_queryset_by_search_term, selection_to_subjects_dict

# create dicts with labels and option values for Select tab
CATEGORY_FILTER_CHOICES = {'all': 'All categories',
                           **{cc[0]: cc[1] + " only" for cc in Surface.CATEGORY_CHOICES}}
SHARING_STATUS_FILTER_CHOICES = {
    'all': 'All accessible datasets',
    'own': 'Only my datasets',
    'shared_ingress': 'Only datasets shared with you',
    'published_ingress': 'Only datasets published by others',
    'shared_egress': 'Only datasets shared by you',
    'published_egress': 'Only datasets published by you'
}
TREE_MODE_CHOICES = ['surface list', 'tag tree']

MAX_PAGE_SIZE = 100
DEFAULT_PAGE_SIZE = 10

DEFAULT_SELECT_TAB_STATE = {
    'search_term': '',  # empty string means: no search
    'category': 'all',
    'sharing_status': 'all',
    'tree_mode': 'surface list',
    'page_size': 10,
    'current_page': 1,
    # all these values are the default if no filter has been applied
    # and the page is loaded the first time
}

DEFAULT_CONTAINER_FILENAME = "digital_surface_twin.zip"

_log = logging.getLogger(__name__)


def filtered_surfaces(request):
    """Return queryset with surfaces matching all filter criteria.

    Surfaces should be
    - readable by the current user
    - filtered by category and sharing status
    - filtered by search expression, if given

    Parameters
    ----------
    request
        Request instance

    Returns
    -------
        Filtered queryset of surfaces
    """

    user = request.user
    # start with all surfaces which are visible for the user
    qs = surfaces_for_user(user)

    #
    # Filter by category and sharing status
    #
    category = get_category(request)
    if category != 'all':
        qs = qs.filter(category=category)

    sharing_status = get_sharing_status(request)

    match sharing_status:
        case 'own':
            qs = qs.filter(creator=user)
        case 'shared_ingress':
            qs = qs.filter(~Q(creator=user))
            if hasattr(Surface, 'publication'):
                qs = qs.exclude(publication__isnull=False)  # exclude published and own surfaces
        case 'published_ingress':
            if hasattr(Surface, 'publication'):
                qs = qs.exclude(publication__isnull=True)
            qs = qs.exclude(creator=user)  # exclude unpublished and own surfaces
        case 'shared_egress':
            PermissionModel = get_user_obj_perms_model(Surface)
            viewable_surfaces_perms = (PermissionModel.objects
                                       .filter(content_object__creator=user,
                                               permission__codename='view_surface')  # only view permissions
                                       .exclude(user=user))  # not own permissions
            qs = qs.filter(id__in=viewable_surfaces_perms.values_list('content_object', flat=True))
            if hasattr(Surface, 'publication'):
                qs = qs.exclude(publication__isnull=False)  # own surfaces, shared with others, unpublished
        case 'published_egress':
            if hasattr(Surface, 'publication'):
                qs = qs.filter(publication__isnull=False)
            qs = qs.filter(creator=user)
        case 'all':
            pass
    #
    # Filter by search term
    #
    search_term = get_search_term(request)
    if search_term:
        #
        # search specific fields of all surfaces in a 'websearch' manner:
        # combine phrases by "AND", allow expressions and quotes
        #
        # See https://docs.djangoproject.com/en/3.2/ref/contrib/postgres/search/#full-text-search
        # for details.
        #
        # We introduce an extra field for search in tag names where the tag names
        # are changed so that the tokenizer splits the names into multiple words
        qs = qs.annotate(
            tag_names_for_search=Replace(
                Replace('tags__name', Value('.'), Value(' ')),  # replace . with space
                Value('/'), Value(' ')),  # replace / with space
            topography_tag_names_for_search=Replace(  # same for the topographies
                Replace('topography__tags__name', Value('.'), Value(' ')),
                Value('/'), Value(' ')),
            topography_name_for_search=Replace('topography__name', Value('.'), Value(' '), output_field=TextField())
            # often there are filenames
        ).distinct('id').order_by('id')
        qs = filter_queryset_by_search_term(qs, search_term, [
            'description', 'name', 'creator__name', 'tag_names_for_search',
            'topography_name_for_search', 'topography__description', 'topography_tag_names_for_search',
            'topography__creator__name',
        ])
    return qs


def download_selection_as_surfaces(request):
    """Returns a file comprised from surfaces related to the selection.

    :param request: current request
    :return:
    """

    from .utils import current_selection_as_surface_list
    surfaces = current_selection_as_surface_list(request)

    container_bytes = BytesIO()
    write_surface_container(container_bytes, surfaces)

    # Prepare response object.
    response = HttpResponse(container_bytes.getvalue(),
                            content_type='application/x-zip-compressed')
    response['Content-Disposition'] = 'attachment; filename="{}"'.format(DEFAULT_CONTAINER_FILENAME)
    # Since the selection contains multiple surfaces in general, we should think about
    # another file name in this case.

    # increase download count for each surface
    for surf in surfaces:
        increase_statistics_by_date_and_object(Metric.objects.SURFACE_DOWNLOAD_COUNT,
                                               period=Period.DAY, obj=surf)

    return response


class TopographyDetailView(TemplateView):
    template_name = "manager/topography_detail.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Get surface instance
        topography_id = self.request.GET.get('topography')
        if topography_id is None:
            return context
        topography = Topography.objects.get(id=int(topography_id))

        #
        # Add context needed for tabs
        #
        context['extra_tabs'] = [
            {
                'title': f"{topography.surface.label}",
                'icon': "gem",
                'icon_style_prefix': 'far',
                'href': f"{reverse('ce_ui:surface-detail')}?surface={topography.surface.pk}",
                'active': False,
                'login_required': False,
                'tooltip': f"Properties of surface '{topography.surface.label}'"
            },
            {
                'title': f"{topography.name}",
                'icon': "file",
                'icon_style_prefix': 'far',
                'href': self.request.path,
                'active': True,
                'login_required': False,
                'tooltip': f"Properties of topography '{topography.name}'"
            }
        ]

        return context


class DataSetListView(TemplateView):
    template_name = "manager/select.html"

    def dispatch(self, request, *args, **kwargs):
        # count this view event for statistics
        metric = Metric.objects.SEARCH_VIEW_COUNT
        increase_statistics_by_date(metric, period=Period.DAY)
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        session = self.request.session

        search_term = get_search_term(self.request)
        if search_term:
            # When searching, we want the default select tab state except for
            # the search term, which is taken from thr request parameters.
            # If not using the default select tab state, this can result
            # in "Load Error!" on the page (#543) because e.g. page 2
            # is not available in the result.
            select_tab_state = DEFAULT_SELECT_TAB_STATE.copy()
            select_tab_state['search_term'] = search_term
        else:
            # .. otherwise keep search term from session variable 'select_tab_state'
            #    and all other state settings
            select_tab_state = session.get('select_tab_state',
                                           default=DEFAULT_SELECT_TAB_STATE.copy())

        # key: tree mode
        context['base_urls'] = {
            'surface list': self.request.build_absolute_uri(reverse('ce_ui:search')),
            'tag tree': self.request.build_absolute_uri(reverse('ce_ui:tag-list')),
        }

        context['category_filter_choices'] = CATEGORY_FILTER_CHOICES.copy()

        if self.request.user.is_anonymous:
            # Anonymous user have only one choice
            context['sharing_status_filter_choices'] = {
                'published_ingress': SHARING_STATUS_FILTER_CHOICES['published_ingress']
            }
            select_tab_state['sharing_status'] = 'published_ingress'  # this only choice should be selected
        else:
            context['sharing_status_filter_choices'] = SHARING_STATUS_FILTER_CHOICES.copy()

        context['select_tab_state'] = select_tab_state.copy()

        # The session needs a default for the state of the select tab
        session['select_tab_state'] = select_tab_state

        return context


class SurfaceDetailView(TemplateView):
    template_name = "manager/surface_detail.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # Get surface instance
        surface_id = self.request.GET.get('surface')
        if surface_id is None:
            return context
        surface = Surface.objects.get(id=int(surface_id))

        context['extra_tabs'] = [
            {
                'title': f"{surface.label}",
                'icon': "gem",
                'icon_style_prefix': 'far',
                'href': f"{reverse('ce_ui:surface-detail')}?surface={surface.pk}",
                'active': False,
                'tooltip': f"Properties of surface '{surface.label}'"
            }
        ]

        return context


#######################################################################################
# Views for REST interface
#######################################################################################
class SurfaceSearchPaginator(PageNumberPagination):
    page_size = DEFAULT_PAGE_SIZE
    page_query_param = 'page'
    page_size_query_param = 'page_size'
    max_page_size = MAX_PAGE_SIZE

    def get_paginated_response(self, data):

        #
        # Save information about requested data in session
        #
        session = self.request.session

        select_tab_state = session.get('select_tab_state', DEFAULT_SELECT_TAB_STATE.copy())
        # not using the keyword argument "default" here, because in some tests,
        # the session is a simple dict and no real session dict. A simple
        # dict's .get() has no keyword argument 'default', although it can be given
        # as second parameter.

        select_tab_state['search_term'] = get_search_term(self.request)
        select_tab_state['category'] = get_category(self.request)
        select_tab_state['sharing_status'] = get_sharing_status(self.request)
        select_tab_state['tree_mode'] = get_tree_mode(self.request)
        page_size = self.get_page_size(self.request)
        select_tab_state[self.page_size_query_param] = page_size
        select_tab_state['current_page'] = self.page.number
        _log.debug("Setting select tab state set in paginator: %s", select_tab_state)
        session['select_tab_state'] = select_tab_state

        return Response({
            'num_items': self.page.paginator.count,
            'num_pages': self.page.paginator.num_pages,
            'page_range': list(self.page.paginator.page_range),
            'page_urls': list(self.get_page_urls()),
            'current_page': self.page.number,
            'num_items_on_current_page': len(self.page.object_list),
            'page_size': page_size,
            'search_term': select_tab_state['search_term'],
            'category': select_tab_state['category'],
            'sharing_status': select_tab_state['sharing_status'],
            'tree_mode': select_tab_state['tree_mode'],
            'page_results': data
        })

    def get_page_urls(self):
        base_url = self.request.build_absolute_uri()
        urls = []
        for page_no in self.page.paginator.page_range:
            if page_no == 1:
                url = remove_query_param(base_url, self.page_query_param)
            else:
                url = replace_query_param(base_url, self.page_query_param, page_no)
            # always add page size, so requests for other pages have it
            url = replace_query_param(url, self.page_size_query_param, self.get_page_size(self.request))
            urls.append(url)
        return urls


class TagTreeView(generics.ListAPIView):
    """
    Generate tree of tags with surfaces and topographies underneath.
    """
    serializer_class = TagSearchSerizalizer
    pagination_class = SurfaceSearchPaginator

    def get_queryset(self):
        surfaces = filtered_surfaces(self.request)
        topographies = filtered_topographies(self.request, surfaces)
        return tags_for_user(self.request.user, surfaces, topographies).filter(parent=None)
        # Only top level are collected, the children are added in the serializer.
        #
        # TODO The filtered surfaces and topographies are calculated twice here, not sure how to circumvent this.
        # Maybe by caching with request argument?

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['selected_instances'] = selected_instances(self.request)
        context['request'] = self.request

        surfaces = filtered_surfaces(self.request)
        topographies = filtered_topographies(self.request, surfaces)
        tags = tags_for_user(self.request.user, surfaces, topographies)
        context['tags_for_user'] = tags

        #
        # also pass filtered surfaces and topographies the user has access to
        #
        context['surfaces'] = surfaces
        context['topographies'] = topographies

        return context


# FIXME!!! This should be folded into the `SurfaceViewSet`, but handling
#  selections should be moved to the client first.
class SurfaceListView(generics.ListAPIView):
    """
    List all surfaces with topographies underneath.
    """
    serializer_class = SurfaceSearchSerializer
    pagination_class = SurfaceSearchPaginator

    def get_queryset(self):
        return filtered_surfaces(self.request)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['selected_instances'] = selected_instances(self.request)
        context['request'] = self.request
        return context


def _selection_set(request):
    return set(request.session.get('selection', []))


def _surface_key(pk):  # TODO use such a function everywhere: instance_key_for_selection()
    return 'surface-{}'.format(pk)


def _topography_key(pk):
    return 'topography-{}'.format(pk)


def _tag_key(pk):
    return 'tag-{}'.format(pk)


def set_surface_select_status(request, pk, select_status):
    """Marks the given surface as 'selected' in session or checks this.

        :param request: request
        :param pk: primary key of the surface
        :param select_status: True if surface should be selected, False if it should be unselected
        :return: JSON Response

        The response returns the current selection as suitable for the basket.
    """
    try:
        pk = int(pk)
        surface = Surface.objects.get(pk=pk)
        assert request.user.has_perm('view_surface', surface)
    except (ValueError, Surface.DoesNotExist, AssertionError):
        raise PermissionDenied()  # This should be shown independent of whether the surface exists

    surface_key = _surface_key(pk)
    selection = _selection_set(request)
    is_selected = surface_key in selection

    if request.method == 'POST':
        if select_status:
            # surface should be selected
            selection.add(surface_key)
        elif is_selected:
            selection.remove(surface_key)

        request.session['selection'] = list(selection)

    data = current_selection_as_basket_items(request)
    return Response(data)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def select_surface(request, pk):
    """Marks the given surface as 'selected' in session.

    :param request: request
    :param pk: primary key of the surface
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    return set_surface_select_status(request, pk, True)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def unselect_surface(request, pk):
    """Marks the given surface as 'unselected' in session.

    :param request: request
    :param pk: primary key of the surface
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    return set_surface_select_status(request, pk, False)


def set_topography_select_status(request, pk, select_status):
    """Marks the given topography as 'selected' or 'unselected' in session.

    :param request: request
    :param pk: primary key of the surface
    :param select_status: True or False, True means "mark as selected", False means "mark as unselected"
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    try:
        pk = int(pk)
        topo = Topography.objects.get(pk=pk)
        assert request.user.has_perm('view_surface', topo.surface)
    except (ValueError, Topography.DoesNotExist, AssertionError):
        raise PermissionDenied()  # This should be shown independent of whether the surface exists

    topography_key = _topography_key(pk)
    selection = _selection_set(request)
    is_selected = topography_key in selection

    if request.method == 'POST':
        if select_status:
            # topography should be selected
            selection.add(topography_key)
        elif is_selected:
            selection.remove(topography_key)

        request.session['selection'] = list(selection)

    data = current_selection_as_basket_items(request)
    return Response(data)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def select_topography(request, pk):
    """Marks the given topography as 'selected' in session.

    :param request: request
    :param pk: primary key of the surface
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    return set_topography_select_status(request, pk, True)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def unselect_topography(request, pk):
    """Marks the given topography as 'selected' in session.

    :param request: request
    :param pk: primary key of the surface
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    return set_topography_select_status(request, pk, False)


def set_tag_select_status(request, pk, select_status):
    """Marks the given tag as 'selected' in session or checks this.

        :param request: request
        :param pk: primary key of the tag
        :param select_status: True if tag should be selected, False if it should be unselected
        :return: JSON Response

        The response returns the current selection as suitable for the basket.
    """
    try:
        pk = int(pk)
        tag = TagModel.objects.get(pk=pk)
    except ValueError:
        raise PermissionDenied()

    if tag not in tags_for_user(request.user):
        raise PermissionDenied()

    tag_key = _tag_key(pk)
    selection = _selection_set(request)
    is_selected = tag_key in selection

    if request.method == 'POST':
        if select_status:
            # tag should be selected
            selection.add(tag_key)
        elif is_selected:
            selection.remove(tag_key)

        request.session['selection'] = list(selection)

    data = current_selection_as_basket_items(request)
    return Response(data)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def select_tag(request, pk):
    """Marks the given tag as 'selected' in session.

    :param request: request
    :param pk: primary key of the tag
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    return set_tag_select_status(request, pk, True)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def unselect_tag(request, pk):
    """Marks the given tag as 'unselected' in session.

    :param request: request
    :param pk: primary key of the tag
    :return: JSON Response

    The response returns the current selection as suitable for the basket.
    """
    return set_tag_select_status(request, pk, False)


@api_view(['POST'])
@permission_classes([])  # We need to override permissions because the anonymous user has read-only access
def unselect_all(request):
    """Removes all selections from session.

    :param request: request
    :return: empty list as JSON Response
    """
    request.session['selection'] = []
    return Response([])


def extra_tabs_if_single_item_selected(topographies, surfaces):
    """Return contribution to context for opening extra tabs if a single topography/surface is selected.

    Parameters
    ----------
    topographies: list of topographies
        Use here the result of function `utils.selected_instances`.

    surfaces: list of surfaces
        Use here the result of function `utils.selected_instances`.

    Returns
    -------
    Sequence of dicts, each dict corresponds to an extra tab.

    """
    tabs = []

    if len(topographies) == 1 and len(surfaces) == 0:
        # exactly one topography was selected -> show also tabs of topography
        topo = topographies[0]
        tabs.extend([
            {
                'title': f"{topo.surface.label}",
                'icon': "gem",
                'icon_style_prefix': 'far',
                'href': f"{reverse('ce_ui:surface-detail')}?surface={topo.surface.pk}",
                'active': False,
                'login_required': False,
                'tooltip': f"Properties of surface '{topo.surface.label}'",
            },
            {
                'title': f"{topo.name}",
                'icon': "file",
                'icon_style_prefix': 'far',
                'href': f"{reverse('ce_ui:topography-detail')}?topography={topo.pk}",
                'active': False,
                'login_required': False,
                'tooltip': f"Properties of measurement '{topo.name}'",
            }
        ])
    elif len(surfaces) == 1 and all(t.surface == surfaces[0] for t in topographies):
        # exactly one surface was selected -> show also tab of surface
        surface = surfaces[0]
        tabs.append(
            {
                'title': f"{surface.label}",
                'icon': 'gem',
                'icon_style_prefix': 'far',
                'href': f"{reverse('ce_ui:surface-detail')}?surface={surface.pk}",
                'active': False,
                'login_required': False,
                'tooltip': f"Properties of surface '{surface.label}'",
            }
        )
    return tabs


class AnalysisResultDetailView(DetailView):
    """Show analyses for a given analysis function.
    """
    model = AnalysisFunction
    template_name = "analysis/analyses_detail.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        function = self.object
        # Check if user is allowed to use this function
        reg = AnalysisRegistry()
        if function.name not in reg.get_analysis_function_names(self.request.user):
            raise PermissionDenied()

        # filter subjects to those this user is allowed to see
        effective_topographies, effective_surfaces, subjects = selection_to_subjects_dict(self.request)

        # get analysis result type
        visualization_app_name, visualization_type = reg.get_visualization_type_for_function_name(function.name)

        context['function'] = function
        context['visualization_type'] = visualization_type

        # Decide whether to open extra tabs for surface/topography details
        tabs = extra_tabs_if_single_item_selected(effective_topographies, effective_surfaces)
        tabs.extend([
            {
                'title': "Analyze",
                'icon': "chart-area",
                'href': f"{reverse('ce_ui:results-list')}?subjects={self.request.GET.get('subjects')}",
                'active': False,
                'login_required': False,
                'tooltip': "Results for selected analysis functions"
            },
            {
                'title': f"{function.name}",
                'icon': "chart-area",
                'href': f"{self.request.path}?subjects={self.request.GET.get('subjects')}",
                'active': True,
                'login_required': False,
                'tooltip': f"Results for analysis '{function.name}'",
                'show_basket': True,
            }
        ])
        context['extra_tabs'] = tabs

        return context


class AnalysesResultListView(TemplateView):
    """View showing analyses from multiple functions.
    """
    template_name = "analysis/analyses_list.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        topographies = []
        surfaces = []

        # Find out what the subjects are. The usual query is a base64 encoded
        # subjects dictionary passed as the 'subjects' argument.
        subjects = self.request.GET.get('subjects')
        topography = self.request.GET.get('topography')
        surface = self.request.GET.get('surface')
        if subjects is not None:
            try:
                subjects = subjects_from_base64(subjects)
            except:  # noqa: E722
                subjects = None

            if subjects is not None:
                # Update session to reflect selection
                topographies = [t for t in subjects if isinstance(t, Topography)]
                surfaces = [t for t in subjects if isinstance(t, Surface)]
                self.request.session['selection'] = instances_to_selection(topographies=topographies, surfaces=surfaces)
        elif topography is not None:
            pass
        elif surface is not None:
            pass

        # Decide whether to open extra tabs for surface/topography details
        tabs = extra_tabs_if_single_item_selected(topographies, surfaces)
        tabs.append({
            'title': "Analyze",
            'icon': "chart-area",
            'icon-style-prefix': 'fas',
            'href': f"{reverse('ce_ui:results-list')}?subjects={self.request.GET.get('subjects')}",
            'active': True,
            'login_required': False,
            'tooltip': "Results for selected analysis functions",
            'show_basket': True,
        })
        context['extra_tabs'] = tabs

        return context


class HomeView(TemplateView):
    template_name = 'pages/home.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        user = self.request.user
        if user.is_anonymous:
            anon = guardian_user_model().get_anonymous()
            context['num_users'] = User.objects.filter(Q(is_active=True) & ~Q(pk=anon.pk)).count()

            current_stats = current_statistics()
        else:
            current_stats = current_statistics(user)

            # count surfaces you can view, but you are not creator
            context['num_shared_surfaces'] = get_objects_for_user(user, 'view_surface', klass=Surface) \
                .filter(~Q(creator=user)).count()

        context['num_surfaces'] = current_stats['num_surfaces_excluding_publications']
        context['num_topographies'] = current_stats['num_topographies_excluding_publications']
        context['num_analyses'] = current_stats['num_analyses_excluding_publications']

        return context


class TermsView(TemplateView):
    template_name = 'pages/termsconditions.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        active_terms = TermsAndConditions.get_active_terms_list()

        if not self.request.user.is_anonymous:
            context['agreed_terms'] = TermsAndConditions.objects.filter(
                userterms__date_accepted__isnull=False,
                userterms__user=self.request.user).order_by('date_created')

            context['not_agreed_terms'] = active_terms.filter(
                Q(userterms=None) |
                (Q(userterms__date_accepted__isnull=True) &
                 Q(userterms__user=self.request.user))).order_by('date_created')

        else:
            context['active_terms'] = active_terms.order_by('date_created')

        context['extra_tabs'] = [{
            'login_required': False,
            'icon': 'file-contract',
            'title': "Terms and Conditions",
            'active': True,
        }]
        context['connect_fixed_tabs_with_extra_tabs'] = False

        return context


class HelpView(TemplateView):
    template_name = 'pages/help.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        context['reader_infos'] = get_reader_infos()
        context['extra_tabs'] = [
            {
                'icon': 'question-circle',
                'title': "Help",
                'href': self.request.path,
                'active': True,
                'login_required': False,
            }
        ]
        return context


#
# The following two views are overwritten from
# termsandconditions package in order to add context
# for the tabbed interface
#
def tabs_for_terms(terms, request_path):
    if len(terms) == 1:
        tab_title = unescape(f"{terms[0].name} {terms[0].version_number}")  # mimics '|safe' as in original template
    else:
        tab_title = "Terms"  # should not happen in Topobank, but just to be safe

    return [
        {
            'icon': 'file-contract',
            'title': "Terms and Conditions",
            'href': reverse('terms'),
            'active': False,
            'login_required': False,
        },
        {
            'icon': 'file-contract',
            'title': tab_title,
            'href': request_path,
            'active': True,
            'login_required': False,
        }
    ]


class TabbedTermsMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['extra_tabs'] = tabs_for_terms(self.get_terms(self.kwargs), self.request.path)
        context['connect_fixed_tabs_with_extra_tabs'] = False
        return context


class TermsDetailView(TabbedTermsMixin, OrigTermsView):
    pass


class TermsAcceptView(TabbedTermsMixin, AcceptTermsView):
    pass



from django.urls import reverse


def test_api():
    """Test API routes"""
    assert reverse('ce_ui:results-detail', kwargs=dict(pk=123)) == '/ui/html/analysis-detail/123/'
    assert reverse('ce_ui:results-list') == '/ui/html/analysis-list/'



from django.core.management.base import BaseCommand
from django.utils import timezone
from django.utils.html import escape

import datetime
import argparse
import decimal
import markdown2

from termsandconditions.models import TermsAndConditions


def valid_date(s):
    try:
        return datetime.strptime(s, "%Y-%m-%d")
    except ValueError:
        msg = "Not a valid date: '{0}'.".format(s)
        raise argparse.ArgumentTypeError(msg)


def valid_version(s):
    try:
        return decimal.Decimal(s)
    except Exception:
        msg = "Not a valid decimal number: '{0}'.".format(s)
        raise argparse.ArgumentTypeError(msg)


class Command(BaseCommand):
    help = """Imports or updates terms and conditions to the database.
    """

    def add_arguments(self, parser):
        parser.add_argument('slug', type=str,
                            help='Slug used for these terms in URLs, e.g. "site-terms".')
        parser.add_argument('terms_version', type=valid_version,
                            help="Version number of these terms.")
        parser.add_argument('terms_file', type=open,
                            help="Path of input file containing the terms in markdown format.")

        parser.add_argument(
            '-a',
            '--date-active',
            type=valid_date,
            dest='date_active',
            help='Date from which onwards these terms should be active.',
        )

        parser.add_argument(
            '-n',
            '--name',
            type=str,
            dest='name',
            help='Name of these terms and conditions. If not given, will be extracted from first line.',
        )

        parser.add_argument(
            '--summary',
            type=str,
            dest='summary',
            help='Short optional text with summary of changes.',
        )

    def handle(self, *args, **options):

        #
        # Read terms and convert to markdown
        #
        terms_file = options['terms_file']
        terms_markdown = terms_file.read()

        #
        # Handle optional arguments
        #
        if options['name']:
            name = options['name'].strip()
        else:
            terms_markdown_lines = terms_markdown.splitlines()
            name = terms_markdown_lines[0].replace('#', ' ').strip()
            terms_markdown = "\n".join(terms_markdown_lines[1:])  # remove first line, will be inserted by terms pkg
            self.stdout.write(self.style.NOTICE("Using first line of terms as name."))

        name = escape(name)  # e.g. convert & to &amp;

        terms_html = markdown2.markdown(terms_markdown)

        now = timezone.now()

        if options['date_active']:
            date_active = options['date_active']
        else:
            date_active = now

        if options['summary']:
            info = options['summary']
        else:
            info = ""  # no longer allowed to be None with migration 0004 in official repo of django-termsandconditons

        #
        # Finally create terms and conditions in database
        #
        TermsAndConditions.objects.create(
            slug=options['slug'].strip(),
            version_number=options['terms_version'],
            name=name,
            date_created=timezone.now(),
            date_active=date_active,
            text=terms_html,
            info=info
        )

        self.stdout.write(self.style.SUCCESS(f"Successfully created terms and conditions '{name}' in database."))



import datetime
import yaml
import zipfile
from io import BytesIO

import pytest

from django.shortcuts import reverse
from django.conf import settings
from rest_framework.test import APIRequestFactory

from trackstats.models import Metric, Period

from topobank.manager.tests.utils import SurfaceFactory, Topography1DFactory, Topography2DFactory, UserFactory

from ..views import DEFAULT_CONTAINER_FILENAME


@pytest.mark.django_db
def test_download_selection(client, mocker, handle_usage_statistics):
    record_mock = mocker.patch('trackstats.models.StatisticByDateAndObject.objects.record')

    user = UserFactory()
    surface1 = SurfaceFactory(creator=user)
    surface2 = SurfaceFactory(creator=user)
    topo1a = Topography1DFactory(surface=surface1)
    Topography2DFactory(surface=surface1)
    Topography1DFactory(surface=surface2)

    factory = APIRequestFactory()

    request = factory.get(reverse('ce_ui:download-selection'))
    request.user = user
    request.session = {
        'selection': [f'topography-{topo1a.id}', f'surface-{surface2.id}']
    }
    from ..views import download_selection_as_surfaces
    response = download_selection_as_surfaces(request)
    assert response.status_code == 200, response.reason
    assert response['Content-Disposition'] == f'attachment; filename="{DEFAULT_CONTAINER_FILENAME}"'

    # open zip file and look into meta file, there should be two surfaces and three topographies
    with zipfile.ZipFile(BytesIO(response.content)) as zf:
        meta_file = zf.open('meta.yml')
        meta = yaml.safe_load(meta_file)
        assert len(meta['surfaces']) == 2
        assert len(meta['surfaces'][0]['topographies']) == 2
        assert len(meta['surfaces'][1]['topographies']) == 1

    # each downloaded surface is counted once
    metric = Metric.objects.SURFACE_DOWNLOAD_COUNT
    today = datetime.date.today()
    if settings.ENABLE_USAGE_STATS:
        record_mock.assert_any_call(metric=metric, object=surface1, period=Period.DAY,
                                    value=1, date=today)
        record_mock.assert_any_call(metric=metric, object=surface2, period=Period.DAY,
                                    value=1, date=today)


#######################################################################
# Selections
#######################################################################

@pytest.mark.django_db
def test_empty_surface_selection(client, handle_usage_statistics):
    #
    # database objects
    #
    user = UserFactory()
    surface = SurfaceFactory(creator=user)
    assert surface.topography_set.count() == 0

    client.force_login(user)

    client.post(reverse('ce_ui:surface-select', kwargs=dict(pk=surface.pk)))

    #
    # Now the selection should contain one empty surface
    #
    assert client.session['selection'] == [f'surface-{surface.pk}']



from django.shortcuts import reverse

from topobank.manager.tests.utils import ordereddicts_to_dicts

from ..views import SurfaceListView


def search_surfaces(request_factory, user, expr):
    """Search surfaces with given expression and return dicts with results.

    This is a helper function used in tests.

    Parameters
    ----------
    request_factory: rest_framework.test.APIRequestFactory
        Used to generate a search request
    user: topobank.users.models.User
        This is the user who performs the search.
    expr: str
        Search expression.

    Returns
    -------
    List of dicts with search results, sorted by 'title' key.
    """
    request = request_factory.get(reverse('ce_ui:search') + f"?search={expr}")
    request.user = user
    request.session = {}  # must be there
    response = SurfaceListView.as_view()(request)
    assert response.status_code == 200
    return ordereddicts_to_dicts(response.data['page_results'], sorted_by='title')



import pytest
from django.shortcuts import reverse

from topobank.manager.models import Surface, Topography
from topobank.manager.tests.utils import Topography1DFactory, Topography2DFactory, SurfaceFactory, TagModelFactory, \
    UserFactory

from ..utils import selection_to_instances, instances_to_selection, tags_for_user, \
    instances_to_topographies, instances_to_surfaces, current_selection_as_surface_list, surface_collection_name


@pytest.fixture
def mock_topos(mocker):
    mocker.patch('topobank.manager.models.Topography', autospec=True)
    mocker.patch('topobank.manager.models.Surface', autospec=True)
    mocker.patch('topobank.manager.models.TagModel', autospec=True)


@pytest.fixture
def testuser(django_user_model):
    username = 'testuser'
    user, created = django_user_model.objects.get_or_create(username=username)
    return user


@pytest.mark.skip('Mocking does not seem to work properly here')
def test_selection_to_instances(testuser, mock_topos):
    from topobank.manager.models import Topography, Surface, TagModel

    selection = ('topography-1', 'topography-2', 'surface-1', 'surface-3', 'tag-1', 'tag-2', 'tag-4')
    selection_to_instances(selection)

    Topography.objects.filter.assert_called_with(id__in={1, 2})
    Surface.objects.filter.assert_called_with(id__in={1, 3})  # set instead of list
    TagModel.objects.filter.assert_called_with(id__in={1, 2, 4})  # set instead of list


@pytest.mark.django_db
def test_instances_to_selection():
    user = UserFactory()

    tag1 = TagModelFactory()
    tag2 = TagModelFactory()

    surface1 = SurfaceFactory(creator=user, tags=[tag1, tag2])
    surface2 = SurfaceFactory(creator=user)

    topo1 = Topography2DFactory(surface=surface1)
    topo2 = Topography1DFactory(surface=surface2, tags=[tag2])

    assert topo1.surface != topo2.surface

    s = instances_to_selection(topographies=[topo1, topo2])

    assert s == [f'topography-{topo1.id}', f'topography-{topo2.id}']

    #
    # It should be possible to give surfaces
    #
    s = instances_to_selection(surfaces=[surface1, surface2])
    assert s == [f'surface-{surface1.id}', f'surface-{surface2.id}']

    #
    # It should be possible to give surfaces and topographies
    #
    s = instances_to_selection(topographies=[topo1], surfaces=[surface2])
    assert s == [f'surface-{surface2.id}', f'topography-{topo1.id}']

    #
    # It should be possible to pass tags
    #
    s = instances_to_selection(tags=[tag2, tag1])
    assert s == [f'tag-{tag1.id}', f'tag-{tag2.id}']

    # Also mixed with other instances
    s = instances_to_selection(tags=[tag2, tag1], topographies=[topo1], surfaces=[surface2])
    assert s == [f'surface-{surface2.id}', f'tag-{tag1.id}', f'tag-{tag2.id}', f'topography-{topo1.id}']


@pytest.mark.django_db
def test_tags_for_user(two_topos):
    topo1 = Topography.objects.get(name="Example 3 - ZSensor")
    topo1.tags = ['rough', 'projects/a']
    topo1.save()

    from topobank.manager.models import TagModel
    print("Tags of topo1:", topo1.tags)
    print(TagModel.objects.all())

    topo2 = Topography.objects.get(name="Example 4 - Default")
    topo2.tags = ['interesting']
    topo2.save()

    surface1 = Surface.objects.get(name="Surface 1")
    surface1.tags = ['projects/C', 'rare']
    surface1.save()

    surface2 = Surface.objects.get(name="Surface 2")
    surface2.tags = ['projects/B', 'a long tag with spaces']
    surface2.save()

    user = surface1.creator

    assert surface2.creator == user

    tags = tags_for_user(user)

    assert set(t.name for t in tags) == {'a long tag with spaces', 'interesting', 'rare', 'rough',
                                         'projects/a', 'projects/b', 'projects/c', 'projects'}


@pytest.mark.django_db
def test_instances_to_topographies(user_three_topographies_three_surfaces_three_tags):
    #
    # Define instances as local variables
    #
    user, (topo1a, topo1b, topo2a), (surface1, surface2, surface3), (tag1, tag2, tag3) \
        = user_three_topographies_three_surfaces_three_tags

    # nothing given, nothing returned
    assert list(instances_to_topographies([], [], [])) == []

    # surface without topographies is the same
    assert list(instances_to_topographies([], [surface3], [])) == []

    # only one surface given
    assert list(instances_to_topographies([], [surface1], [])) == [topo1a, topo1b]

    # only two surfaces given
    assert list(instances_to_topographies([], [surface2, surface1], [])) == [topo1a, topo1b, topo2a]

    # an empty surface makes no difference here
    assert list(instances_to_topographies([], [surface2, surface1, surface3], [])) == [topo1a, topo1b, topo2a]

    # an additional topography makes no difference
    assert list(instances_to_topographies([topo1a], [surface1], [])) == [topo1a, topo1b]

    # also single topographies can be selected
    assert list(instances_to_topographies([topo2a, topo1b], [], [])) == [topo1b, topo2a]

    # a single tag can be selected
    assert list(instances_to_topographies([], [], [tag3])) == [topo1b]

    # an additional topography given does not change result if already tagged the same way
    assert list(instances_to_topographies([topo1b], [], [tag3])) == [topo1b]

    # also two tags can be given
    assert list(instances_to_topographies([], [], [tag2, tag3])) == [topo1b, topo2a]


@pytest.mark.django_db
def test_instances_to_surfaces(user_three_topographies_three_surfaces_three_tags):
    #
    # Define instances as local variables
    #
    user, (topo1a, topo1b, topo2a), (surface1, surface2, surface3), (tag1, tag2, tag3) \
        = user_three_topographies_three_surfaces_three_tags

    # nothing given, nothing returned
    assert list(instances_to_surfaces([], [])) == []

    # surface without topographies is the same
    assert list(instances_to_surfaces([surface3], [])) == [surface3]

    # two surfaces given
    assert list(instances_to_surfaces([surface2, surface1], [])) == [surface1, surface2]

    # a single tag can be selected
    assert list(instances_to_surfaces([], [tag3])) == [surface3]

    # also two tags can be given
    assert list(instances_to_surfaces([], [tag2, tag3])) == [surface2, surface3]


@pytest.mark.django_db
def test_related_surfaces_for_selection(rf):
    user = UserFactory()

    # create tags
    tag1 = TagModelFactory(name='apple')
    tag2 = TagModelFactory(name='banana')

    # create surfaces
    surf1 = SurfaceFactory(creator=user, tags=[tag1])
    surf2 = SurfaceFactory(creator=user)
    surf3 = SurfaceFactory(creator=user, tags=[tag1])

    # add some topographies
    topo1a = Topography1DFactory(surface=surf1)
    Topography2DFactory(surface=surf1, tags=[tag2])
    Topography2DFactory(surface=surf2, tags=[tag1])

    # surf3 has no topography

    def get_request(topographies=[], surfaces=[], tags=[]):
        """Simple get request while setting the selection"""
        request = rf.get(reverse('ce_ui:select'))
        request.user = user
        request.session = {'selection': instances_to_selection(
            topographies=topographies,
            surfaces=surfaces,
            tags=tags,
        )}
        return request

    # tag 'apple' should return all three surfaces
    assert current_selection_as_surface_list(get_request(tags=[tag1])) == [surf1, surf2, surf3]

    # one topography should return its surface
    assert current_selection_as_surface_list(get_request(topographies=[topo1a])) == [surf1]

    # We should be able to mix tags, topographies and surfaces
    assert current_selection_as_surface_list(get_request(topographies=[topo1a],
                                                         surfaces=[surf1],
                                                         tags=[tag2])) == [surf1]


@pytest.mark.parametrize(["surface_names", "exp_name", "max_total_length"], [
    (["A"], "Surface 'A'", 11),
    (["A", "B"], "Surface 'A', Surface 'B'", 30),
    (["AAAAA"], "Surface 'AAAAA'", 15),
    (["AAAAAA"], "Surface 'A...", 14),
    (["AAA", "BBB", "CCCCCCCCCCCCCCCCCCCCCCCC"],
     "Surface 'AAA', Surface 'BBB' and 1 more", 40),
])
def test_surface_collection_name(surface_names, exp_name, max_total_length):
    assert surface_collection_name(surface_names, max_total_length=max_total_length) == exp_name



from django.conf import settings
from django.urls import include, re_path, path
from django.views.generic import TemplateView, RedirectView

from . import views

app_name = "ce_ui"

urlprefix = None  # No url prefix, this plugin wants to register top-level routes

#
# Top-level routes
#
urlpatterns = [
    #
    # Main entry points and static pages
    #
    path("", views.HomeView.as_view(), name="home"),
    path(
        "about/",
        TemplateView.as_view(template_name="pages/about.html"),
        name="about",
    ),
    path(
        "termsandconditions/",
        views.TermsView.as_view(),
        name="terms",
    ),
    path(
        "help/",
        views.HelpView.as_view(),
        name="help",
    ),
    path(
        "search/",
        RedirectView.as_view(pattern_name='ce_ui:select'),
        name="search",
    ),
    #
    # For asking for terms and conditions
    #
    # some url specs are overwritten here pointing to own views in order to plug in
    # some extra context for the tabbed interface
    # View Specific Active Terms
    re_path(r'^terms/view/(?P<slug>[a-zA-Z0-9_.-]+)/$', views.TermsDetailView.as_view(),
            name="tc_view_specific_page"),

    # View Specific Version of Terms
    re_path(r'^terms/view/(?P<slug>[a-zA-Z0-9_.-]+)/(?P<version>[0-9.]+)/$', views.TermsDetailView.as_view(),
            name="tc_view_specific_version_page"),

    # Print Specific Version of Terms
    re_path(r'^terms/print/(?P<slug>[a-zA-Z0-9_.-]+)/(?P<version>[0-9.]+)/$',
            views.TermsDetailView.as_view(template_name="termsandconditions/tc_print_terms.html"),
            name="tc_print_page"),

    # Accept Terms
    re_path(r'^terms/accept/$', views.TermsAcceptView.as_view(), name="tc_accept_page"),

    # Accept Specific Terms
    re_path(r'^terms/accept/(?P<slug>[a-zA-Z0-9_.-]+)$', views.TermsAcceptView.as_view(),
            name="tc_accept_specific_page"),

    # Accept Specific Terms Version
    re_path(r'^terms/accept/(?P<slug>[a-zA-Z0-9_.-]+)/(?P<version>[0-9\.]+)/$', views.TermsAcceptView.as_view(),
            name="tc_accept_specific_version_page"),

    # the defaults
    re_path(r'^terms/', include('termsandconditions.urls')),
]

#
# Routes under the 'ui/' prefix
#
ui_urlpatterns = [
    #
    # HTML routes
    #
    path(
        'html/dataset-list/',
        view=views.DataSetListView.as_view(),
        name='select'
    ),
    path(
        r'html/topography/',
        view=views.TopographyDetailView.as_view(),
        name='topography-detail'
    ),
    path(
        r'html/surface/',
        view=views.SurfaceDetailView.as_view(),
        name='surface-detail'
    ),
    path(
        'html/analysis-list/',
        view=views.AnalysesResultListView.as_view(),
        name='results-list'
    ),
    path(
        r'html/analysis-detail/<int:pk>/',
        view=views.AnalysisResultDetailView.as_view(),
        name='results-detail'
    ),
    #
    # Data routes
    #
    path(
        'select/download/',
        view=views.download_selection_as_surfaces,
        name='download-selection'
    ),
    #
    # API routes
    #
    path(
        'api/search/',  # TODO check URL, rename?
        view=views.SurfaceListView.as_view(),  # TODO Check view name, rename?
        name='search'  # TODO rename?
    ),
    path(
        'api/tag-tree/',
        view=views.TagTreeView.as_view(),
        name='tag-list'  # TODO rename
    ),
    re_path(
        r'api/selection/surface/(?P<pk>\d+)/select/$',
        view=views.select_surface,
        name='surface-select'
    ),
    re_path(
        r'api/selection/surface/(?P<pk>\d+)/unselect/$',
        view=views.unselect_surface,
        name='surface-unselect'
    ),
    re_path(
        r'api/selection/topography/(?P<pk>\d+)/select/$',
        view=views.select_topography,
        name='topography-select'
    ),
    re_path(
        r'api/selection/topography/(?P<pk>\d+)/unselect/$',
        view=views.unselect_topography,
        name='topography-unselect'
    ),
    re_path(
        r'api/selection/tag/(?P<pk>\d+)/select/$',
        view=views.select_tag,
        name='tag-select'
    ),
    re_path(
        r'api/selection/tag/(?P<pk>\d+)/unselect/$',
        view=views.unselect_tag,
        name='tag-unselect'
    ),
    path(
        'api/selection/unselect-all/',
        view=views.unselect_all,
        name='unselect-all'
    )
]
urlpatterns += [path('ui/', include((ui_urlpatterns, app_name)))]

#
# Challenge redirect
#
if settings.CHALLENGE_REDIRECT_URL:
    urlpatterns += [
        path(
            "challenge/",
            RedirectView.as_view(url=settings.CHALLENGE_REDIRECT_URL, permanent=False),
            name="challenge",
        ),
    ]



import pytest
from django.shortcuts import reverse

from topobank.utils import assert_in_content

from topobank.analysis.tests.utils import TopographyAnalysisFactory
from topobank.manager.tests.utils import SurfaceFactory, Topography1DFactory, UserFactory


@pytest.mark.django_db
@pytest.fixture
def test_instances(test_analysis_function):
    users = [
        UserFactory(username='user1'),
        UserFactory(username='user2')
    ]

    surfaces = [
        SurfaceFactory(creator=users[0]),
        SurfaceFactory(creator=users[0]),
    ]

    topographies = [
        Topography1DFactory(surface=surfaces[0])
    ]

    TopographyAnalysisFactory(function=test_analysis_function, subject_topography=topographies[0])

    return users, surfaces, topographies


@pytest.mark.django_db
def test_welcome_page_statistics(client, test_instances, handle_usage_statistics):
    (user_1, user_2), (surface_1, surface_2), (topography_1,) = test_instances
    surface_2.share(user_2)

    #
    # Test statistics if user is not yet authenticated
    #
    response = client.get(reverse('home'))

    assert_in_content(response, '<div class="welcome-page-statistics">2</div> registered users')
    assert_in_content(response, '<div class="welcome-page-statistics">2</div> digital surface twins')
    assert_in_content(response, '<div class="welcome-page-statistics">1</div> individual measurements')
    assert_in_content(response, '<div class="welcome-page-statistics">1</div> computed analyses')

    #
    # Test statistics if user_1 is authenticated
    #
    client.force_login(user_1)
    response = client.get(reverse('home'))

    assert_in_content(response, '<div class="welcome-page-statistics">2</div> digital surface twins')
    assert_in_content(response, '<div class="welcome-page-statistics">1</div> individual measurements')
    assert_in_content(response, '<div class="welcome-page-statistics">1</div> computed analyses')
    assert_in_content(response, '<div class="welcome-page-statistics">0</div> digital twins of other users')

    client.logout()

    #
    # Test statistics if user_2 is authenticated
    #
    client.force_login(user_2)
    response = client.get(reverse('home'))

    assert_in_content(response, '<div class="welcome-page-statistics">0</div> digital surface twins')
    assert_in_content(response, '<div class="welcome-page-statistics">0</div> individual measurements')
    assert_in_content(response, '<div class="welcome-page-statistics">0</div> computed analyses')
    assert_in_content(response, '<div class="welcome-page-statistics">1</div> digital twins of other users')

    client.logout()



"""Test related to searching"""

import pytest

from django.shortcuts import reverse
from rest_framework.test import APIRequestFactory

from topobank.manager.models import TagModel
from topobank.manager.utils import subjects_to_base64

from topobank.manager.tests.utils import ordereddicts_to_dicts, Topography1DFactory, UserFactory, SurfaceFactory

from ..views import SurfaceSearchPaginator, SurfaceListView, TagTreeView

from .utils import search_surfaces


def assert_dict_equal(a, b, key=None):
    try:
        keys_a = set(a.keys())
        keys_b = set(b.keys())
    except AttributeError:
        assert a == b, f'The value of the following key differs: {key}'
        return

    assert keys_a == keys_b, f'The following keys are not present in both dictionaries: {keys_a ^ keys_b}'
    for key in keys_a:
        if isinstance(a[key], dict):
            assert_dict_equal(a[key], b[key], key=key)
        elif isinstance(a[key], list):
            assert_dicts_equal(a[key], b[key])
        else:
            assert a[key] == b[key], f'The value of the following key differs: {key}'


def assert_dicts_equal(a, b):
    for x, y in zip(a, b):
        assert_dict_equal(x, y)


@pytest.fixture
def user_three_surfaces_four_topographies():
    #
    # Create some database objects
    #
    user = UserFactory()
    surface1 = SurfaceFactory(creator=user, category='exp')
    surface2 = SurfaceFactory(creator=user, category='sim')
    surface3 = SurfaceFactory(creator=user, category='dum')

    topo1a = Topography1DFactory(surface=surface1)
    topo1b = Topography1DFactory(surface=surface1)
    topo2a = Topography1DFactory(surface=surface2)
    topo2b = Topography1DFactory(surface=surface2)
    # no topography for surface3 on purpose

    return user, surface1, surface2, surface3, topo1a, topo1b, topo2a, topo2b


@pytest.mark.django_db
def test_surface_search_with_request_factory(user_three_surfaces_four_topographies):
    user, surface1, surface2, surface3, topo1a, topo1b, topo2a, topo2b = user_three_surfaces_four_topographies

    #
    # Set some tags
    #
    surface1.tags = ['bike', 'train/tgv']
    surface1.save()
    topo2a.tags = ['bike', 'train/ice']
    topo2a.save()

    #
    # Fix a selection and create a request with this selection
    #
    session = dict(selection=[f'surface-{surface2.pk}', f'topography-{topo1a.pk}', f'surface-{surface3.pk}'])

    factory = APIRequestFactory()
    request = factory.get(reverse('ce_ui:search'))  # no search term here, see below for another search with term
    request.user = user
    request.session = session

    #
    # Create search response and compare with expectation
    #
    assert SurfaceSearchPaginator.page_size >= 3  # needed in order to get all these test results
    response = SurfaceListView.as_view()(request)

    assert response.status_code == 200

    user_url = request.build_absolute_uri(user.get_absolute_url())

    topo1a_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo1a])}"
    topo1b_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo1b])}"
    topo2a_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo2a])}"
    topo2b_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo2b])}"
    surface1_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([surface1])}"
    surface2_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([surface2])}"

    expected_dicts = [
        {
            'category': 'exp',
            'category_name': 'Experimental data',
            'children': [
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo1a.pk}',
                 'surface_key': f'surface-{surface1.pk}',
                 'label': topo1a.label,
                 'name': topo1a.name,
                 'id': topo1a.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': True,
                 'sharing_status': 'own',
                 'tags': [],
                 'title': topo1a.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': int(topo1a.height_scale),
                 'height_scale_editable': topo1a.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo1a.is_periodic,
                 'measurement_date': str(topo1a.measurement_date),
                 'resolution_x': topo1a.resolution_x,
                 'resolution_y': topo1a.resolution_y,
                 'size_editable': topo1a.size_editable,
                 'size_x': topo1a.size_x,
                 'size_y': topo1a.size_y,
                 'thumbnail': None,
                 'unit': topo1a.unit,
                 'unit_editable': topo1a.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo1a.id}',
                          'select': f'/ui/api/selection/topography/{topo1a.id}/select/',
                          'analyze': topo1a_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo1a.id}/unselect/'}},
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo1b.pk}',
                 'surface_key': f'surface-{surface1.pk}',
                 'label': topo1b.label,
                 'name': topo1b.name,
                 'id': topo1b.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': False,
                 'sharing_status': 'own',
                 'tags': [],
                 'title': topo1b.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': topo1b.height_scale,
                 'height_scale_editable': topo1b.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo1b.is_periodic,
                 'measurement_date': str(topo1b.measurement_date),
                 'resolution_x': topo1b.resolution_x,
                 'resolution_y': topo1b.resolution_y,
                 'size_editable': topo1b.size_editable,
                 'size_x': topo1b.size_x,
                 'size_y': topo1b.size_y,
                 'thumbnail': None,
                 'unit': topo1b.unit,
                 'unit_editable': topo1b.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo1b.id}',
                          'select': f'/ui/api/selection/topography/{topo1b.id}/select/',
                          'analyze': topo1b_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo1b.id}/unselect/'}},

            ],
            'creator': user_url,
            'creator_name': user.name,
            'description': '',
            'folder': True,
            'key': f'surface-{surface1.pk}',
            'label': surface1.label,
            'name': surface1.name,
            'id': surface1.pk,
            'publication_authors': None,
            'publication_date': None,
            'publication_license': None,
            'publication_doi': None,
            'selected': False,
            'sharing_status': 'own',
            'tags': ['bike', 'train/tgv'],
            'title': surface1.name,
            'topography_count': 2,
            'type': 'surface',
            'version': None,
            'urls': {'analyze': surface1_analyze,
                     'download': f'/manager/api/surface/{surface1.id}/download/',
                     'detail': f'/ui/html/surface/?surface={surface1.id}',
                     'select': f'/ui/api/selection/surface/{surface1.id}/select/',
                     'unselect': f'/ui/api/selection/surface/{surface1.id}/unselect/'}
        },
        {
            'category': 'sim',
            'category_name': 'Simulated data',
            'children': [
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo2a.pk}',
                 'surface_key': f'surface-{surface2.pk}',
                 'label': topo2a.label,
                 'name': topo2a.name,
                 'id': topo2a.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': False,  # not explicitly selected
                 'sharing_status': 'own',
                 'tags': ['bike', 'train/ice'],
                 'title': topo2a.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': topo2a.height_scale,
                 'height_scale_editable': topo2a.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo2a.is_periodic,
                 'measurement_date': str(topo2a.measurement_date),
                 'resolution_x': topo2a.resolution_x,
                 'resolution_y': topo2a.resolution_y,
                 'size_editable': topo2a.size_editable,
                 'size_x': topo2a.size_x,
                 'size_y': topo2a.size_y,
                 'thumbnail': None,
                 'unit': topo2a.unit,
                 'unit_editable': topo2a.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo2a.id}',
                          'select': f'/ui/api/selection/topography/{topo2a.id}/select/',
                          'analyze': topo2a_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo2a.id}/unselect/'}},
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo2b.pk}',
                 'surface_key': f'surface-{surface2.pk}',
                 'label': topo2b.label,
                 'name': topo2b.name,
                 'id': topo2b.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': False,  # not explicitly selected
                 'sharing_status': 'own',
                 'tags': [],
                 'title': topo2b.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': topo2b.height_scale,
                 'height_scale_editable': topo2b.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo2b.is_periodic,
                 'measurement_date': str(topo2b.measurement_date),
                 'resolution_x': topo2b.resolution_x,
                 'resolution_y': topo2b.resolution_y,
                 'size_editable': topo2b.size_editable,
                 'size_x': topo2b.size_x,
                 'size_y': topo2b.size_y,
                 'thumbnail': None,
                 'unit': topo2b.unit,
                 'unit_editable': topo2b.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo2b.id}',
                          'select': f'/ui/api/selection/topography/{topo2b.id}/select/',
                          'analyze': topo2b_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo2b.id}/unselect/'}},
            ],
            'creator': user_url,
            'creator_name': user.name,
            'description': '',
            'folder': True,
            'key': f'surface-{surface2.pk}',
            'label': surface2.label,
            'name': surface2.name,
            'id': surface2.pk,
            'publication_authors': None,
            'publication_date': None,
            'publication_license': None,
            'publication_doi': None,
            'selected': True,
            'sharing_status': 'own',
            'tags': [],
            'title': surface2.name,
            'topography_count': 2,
            'type': 'surface',
            'version': None,
            'urls': {'analyze': surface2_analyze,
                     'download': f'/manager/api/surface/{surface2.id}/download/',
                     'detail': f'/ui/html/surface/?surface={surface2.id}',
                     'select': f'/ui/api/selection/surface/{surface2.id}/select/',
                     'unselect': f'/ui/api/selection/surface/{surface2.id}/unselect/'}
        },
        {
            'category': 'dum',
            'category_name': 'Dummy data',
            'children': [],
            'creator': user_url,
            'creator_name': user.name,
            'description': '',
            'folder': True,
            'key': f'surface-{surface3.pk}',
            'label': surface3.label,
            'name': surface3.name,
            'id': surface3.pk,
            'publication_authors': None,
            'publication_date': None,
            'publication_license': None,
            'publication_doi': None,
            'selected': True,
            'sharing_status': 'own',
            'tags': [],
            'title': surface3.name,
            'topography_count': 0,
            'type': 'surface',
            'version': None,
            'urls': {'download': f'/manager/api/surface/{surface3.id}/download/',
                     'detail': f'/ui/html/surface/?surface={surface3.id}',
                     'select': f'/ui/api/selection/surface/{surface3.id}/select/',
                     'unselect': f'/ui/api/selection/surface/{surface3.id}/unselect/'}
        },
    ]

    assert ordereddicts_to_dicts(response.data['page_results']) == expected_dicts

    #
    # Do a search and check for reduced results because search for "topo2a"
    #
    request = factory.get(reverse('ce_ui:search') + f"?search={topo2a.name}")
    request.user = user
    request.session = session

    #
    # Create search response and compare with expectation
    #
    response = SurfaceListView.as_view()(request)

    assert response.status_code == 200

    expected_dicts = [
        {
            'category': 'sim',
            'category_name': 'Simulated data',
            'children': [
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo2a.pk}',
                 'surface_key': f'surface-{surface2.pk}',
                 'label': topo2a.label,
                 'name': topo2a.name,
                 'id': topo2a.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': False,  # not explicitly selected
                 'sharing_status': 'own',
                 'tags': ['bike', 'train/ice'],
                 'title': topo2a.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': int(topo2a.height_scale),
                 'height_scale_editable': topo2a.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo2a.is_periodic,
                 'measurement_date': str(topo2a.measurement_date),
                 'resolution_x': topo2a.resolution_x,
                 'resolution_y': topo2a.resolution_y,
                 'size_editable': topo2a.size_editable,
                 'size_x': topo2a.size_x,
                 'size_y': topo2a.size_y,
                 'thumbnail': None,
                 'unit': topo2a.unit,
                 'unit_editable': topo2a.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo2a.id}',
                          'select': f'/ui/api/selection/topography/{topo2a.id}/select/',
                          'analyze': topo2a_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo2a.id}/unselect/'}},
            ],
            'creator': user_url,
            'creator_name': user.name,
            'description': '',
            'folder': True,
            'key': f'surface-{surface2.pk}',
            'label': surface2.label,
            'name': surface2.name,
            'id': surface2.pk,
            'publication_authors': None,
            'publication_date': None,
            'publication_license': None,
            'publication_doi': None,
            'selected': True,
            'sharing_status': 'own',
            'tags': [],
            'title': surface2.name,
            'topography_count': 2,
            'type': 'surface',
            'version': None,
            'urls': {'analyze': surface2_analyze,
                     'download': f'/manager/api/surface/{surface2.id}/download/',
                     'detail': f'/ui/html/surface/?surface={surface2.id}',
                     'select': f'/ui/api/selection/surface/{surface2.id}/select/',
                     'unselect': f'/ui/api/selection/surface/{surface2.id}/unselect/'}
        },
    ]

    resulted_dicts = ordereddicts_to_dicts(response.data['page_results'], sorted_by='title')
    assert resulted_dicts == expected_dicts

    #
    # Do a search and check for reduced results because search for category "exp"
    #
    request = factory.get(reverse('ce_ui:search') + "?category=exp")
    request.user = user
    request.session = session

    #
    # Create search response and compare with expectation
    #
    response = SurfaceListView.as_view()(request)

    assert response.status_code == 200

    expected_dicts = [
        {
            'category': 'exp',
            'category_name': 'Experimental data',
            'children': [
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo1a.pk}',
                 'surface_key': f'surface-{surface1.pk}',
                 'label': topo1a.label,
                 'name': topo1a.name,
                 'id': topo1a.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': True,
                 'sharing_status': 'own',
                 'tags': [],
                 'title': topo1a.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': int(topo1a.height_scale),
                 'height_scale_editable': topo1a.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo1a.is_periodic,
                 'measurement_date': str(topo1a.measurement_date),
                 'resolution_x': topo1a.resolution_x,
                 'resolution_y': topo1a.resolution_y,
                 'size_editable': topo1a.size_editable,
                 'size_x': topo1a.size_x,
                 'size_y': topo1a.size_y,
                 'thumbnail': None,
                 'unit': topo1a.unit,
                 'unit_editable': topo1a.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo1a.id}',
                          'select': f'/ui/api/selection/topography/{topo1a.id}/select/',
                          'analyze': topo1a_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo1a.id}/unselect/'}},
                {'creator': user_url,
                 'creator_name': user.name,
                 'description': '',
                 'folder': False,
                 'key': f'topography-{topo1b.pk}',
                 'surface_key': f'surface-{surface1.pk}',
                 'label': topo1b.label,
                 'name': topo1b.name,
                 'id': topo1b.pk,
                 'publication_authors': None,
                 'publication_date': None,
                 'selected': False,
                 'sharing_status': 'own',
                 'tags': [],
                 'title': topo1b.name,
                 'type': 'topography',
                 'version': None,
                 'datafile_format': None,
                 'height_scale': int(topo1b.height_scale),
                 'height_scale_editable': topo1b.height_scale_editable,
                 'instrument_name': '',
                 'instrument_parameters': {},
                 'instrument_type': 'undefined',
                 'is_periodic': topo1b.is_periodic,
                 'measurement_date': str(topo1b.measurement_date),
                 'resolution_x': topo1b.resolution_x,
                 'resolution_y': topo1b.resolution_y,
                 'size_editable': topo1b.size_editable,
                 'size_x': topo1b.size_x,
                 'size_y': topo1b.size_y,
                 'thumbnail': None,
                 'unit': topo1b.unit,
                 'unit_editable': topo1b.unit_editable,
                 'urls': {'detail': f'/ui/html/topography/?topography={topo1b.id}',
                          'select': f'/ui/api/selection/topography/{topo1b.id}/select/',
                          'analyze': topo1b_analyze,
                          'unselect': f'/ui/api/selection/topography/{topo1b.id}/unselect/'}},

            ],
            'creator': user_url,
            'creator_name': user.name,
            'description': '',
            'folder': True,
            'key': f'surface-{surface1.pk}',
            'label': surface1.label,
            'name': surface1.name,
            'id': surface1.pk,
            'publication_authors': None,
            'publication_date': None,
            'publication_license': None,
            'publication_doi': None,
            'selected': False,
            'sharing_status': 'own',
            'tags': ['bike', 'train/tgv'],
            'title': surface1.name,
            'topography_count': 2,
            'type': 'surface',
            'version': None,
            'urls': {'analyze': surface1_analyze,
                     'download': f'/manager/api/surface/{surface1.id}/download/',
                     'detail': f'/ui/html/surface/?surface={surface1.id}',
                     'select': f'/ui/api/selection/surface/{surface1.id}/select/',
                     'unselect': f'/ui/api/selection/surface/{surface1.id}/unselect/'}
        },
    ]

    resulted_dicts = ordereddicts_to_dicts(response.data['page_results'], sorted_by='title')
    assert resulted_dicts == expected_dicts


@pytest.mark.django_db
def test_tag_search_with_request_factory(user_three_surfaces_four_topographies):
    user, surface1, surface2, surface3, topo1a, topo1b, topo2a, topo2b = user_three_surfaces_four_topographies

    #
    # Set some tags
    #
    surface1.tags = ['bike', 'train/tgv']
    surface1.save()
    topo2a.tags = ['bike', 'train/ice']
    topo2a.save()
    topo2b.tags = ['train/ice/restaurant']
    topo2b.save()

    #
    # Fix a selection and create a request with this selection
    #
    session = dict(selection=[f'surface-{surface2.pk}', f'topography-{topo1a.pk}', f'surface-{surface3.pk}'])

    factory = APIRequestFactory()
    request = factory.get(reverse('ce_ui:tag-list'))
    request.user = user
    request.session = session

    #
    # Create tag tree and compare with expectation
    #
    response = TagTreeView.as_view()(request)

    assert response.status_code == 200

    user_url = request.build_absolute_uri(user.get_absolute_url())

    topo1a_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo1a])}"
    topo1b_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo1b])}"
    topo2a_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo2a])}"
    topo2b_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([topo2b])}"
    surface1_analyze = f"/ui/html/analysis-list/?subjects={subjects_to_base64([surface1])}"

    expected_dict_topo1a = {
        'creator': user_url,
        'creator_name': user.name,
        'description': '',
        'folder': False,
        'key': f'topography-{topo1a.pk}',
        'surface_key': f'surface-{surface1.pk}',
        'label': topo1a.label,
        'name': topo1a.name,
        'id': topo1a.pk,
        'publication_authors': None,
        'publication_date': None,
        'selected': True,
        'sharing_status': 'own',
        'tags': [],
        'title': topo1a.name,
        'type': 'topography',
        'version': None,
        'datafile_format': None,
        'height_scale': int(topo1a.height_scale),
        'height_scale_editable': topo1a.height_scale_editable,
        'instrument_name': '',
        'instrument_parameters': {},
        'instrument_type': 'undefined',
        'is_periodic': topo1a.is_periodic,
        'measurement_date': str(topo1a.measurement_date),
        'resolution_x': topo1a.resolution_x,
        'resolution_y': topo1a.resolution_y,
        'size_editable': topo1a.size_editable,
        'size_x': topo1a.size_x,
        'size_y': topo1a.size_y,
        'thumbnail': None,
        'unit': topo1a.unit,
        'unit_editable': topo1a.unit_editable,
        'urls': {'detail': f'/ui/html/topography/?topography={topo1a.id}',
                 'select': f'/ui/api/selection/topography/{topo1a.id}/select/',
                 'analyze': topo1a_analyze,
                 'unselect': f'/ui/api/selection/topography/{topo1a.id}/unselect/'}
    }
    expected_dict_topo1b = {
        'creator': user_url,
        'creator_name': user.name,
        'description': '',
        'folder': False,
        'key': f'topography-{topo1b.pk}',
        'surface_key': f'surface-{surface1.pk}',
        'label': topo1b.label,
        'name': topo1b.name,
        'id': topo1b.pk,
        'publication_authors': None,
        'publication_date': None,
        'selected': False,
        'sharing_status': 'own',
        'tags': [],
        'title': topo1b.name,
        'type': 'topography',
        'version': None,
        'datafile_format': None,
        'height_scale': int(topo1b.height_scale),
        'height_scale_editable': topo1b.height_scale_editable,
        'instrument_name': '',
        'instrument_parameters': {},
        'instrument_type': 'undefined',
        'is_periodic': topo1b.is_periodic,
        'measurement_date': str(topo1b.measurement_date),
        'resolution_x': topo1b.resolution_x,
        'resolution_y': topo1b.resolution_y,
        'size_editable': topo1b.size_editable,
        'size_x': topo1b.size_x,
        'size_y': topo1b.size_y,
        'thumbnail': None,
        'unit': topo1b.unit,
        'unit_editable': topo1b.unit_editable,
        'urls': {'detail': f'/ui/html/topography/?topography={topo1b.id}',
                 'select': f'/ui/api/selection/topography/{topo1b.id}/select/',
                 'analyze': topo1b_analyze,
                 'unselect': f'/ui/api/selection/topography/{topo1b.id}/unselect/'}
    }

    expected_dict_topo2a = {
        'creator': user_url,
        'creator_name': user.name,
        'description': '',
        'folder': False,
        'key': f'topography-{topo2a.pk}',
        'surface_key': f'surface-{surface2.pk}',
        'label': topo2a.label,
        'name': topo2a.name,
        'id': topo2a.pk,
        'publication_authors': None,
        'publication_date': None,
        'selected': False,  # not explicitly selected
        'sharing_status': 'own',
        'tags': ['bike', 'train/ice'],
        'title': topo2a.name,
        'type': 'topography',
        'version': None,
        'datafile_format': None,
        'height_scale': int(topo2a.height_scale),
        'height_scale_editable': topo2a.height_scale_editable,
        'instrument_name': '',
        'instrument_parameters': {},
        'instrument_type': 'undefined',
        'is_periodic': topo2a.is_periodic,
        'measurement_date': str(topo2a.measurement_date),
        'resolution_x': topo2a.resolution_x,
        'resolution_y': topo2a.resolution_y,
        'size_editable': topo2a.size_editable,
        'size_x': topo2a.size_x,
        'size_y': topo2a.size_y,
        'thumbnail': None,
        'unit': topo2a.unit,
        'unit_editable': topo2a.unit_editable,
        'urls': {'detail': f'/ui/html/topography/?topography={topo2a.id}',
                 'select': f'/ui/api/selection/topography/{topo2a.id}/select/',
                 'analyze': topo2a_analyze,
                 'unselect': f'/ui/api/selection/topography/{topo2a.id}/unselect/'}
    }

    expected_dict_topo2b = {
        'creator': user_url,
        'creator_name': user.name,
        'description': '',
        'folder': False,
        'key': f'topography-{topo2b.pk}',
        'surface_key': f'surface-{surface2.pk}',
        'label': topo2b.label,
        'name': topo2b.name,
        'id': topo2b.pk,
        'publication_authors': None,
        'publication_date': None,
        'selected': False,  # not explicitly selected
        'sharing_status': 'own',
        'tags': ['train/ice/restaurant'],
        'title': topo2b.name,
        'type': 'topography',
        'version': None,
        'datafile_format': None,
        'height_scale': int(topo2b.height_scale),
        'height_scale_editable': topo2b.height_scale_editable,
        'instrument_name': '',
        'instrument_parameters': {},
        'instrument_type': 'undefined',
        'is_periodic': topo2b.is_periodic,
        'measurement_date': str(topo2b.measurement_date),
        'resolution_x': topo2b.resolution_x,
        'resolution_y': topo2b.resolution_y,
        'size_editable': topo2b.size_editable,
        'size_x': topo2b.size_x,
        'size_y': topo2b.size_y,
        'thumbnail': None,
        'unit': topo2b.unit,
        'unit_editable': topo2b.unit_editable,
        'urls': {'detail': f'/ui/html/topography/?topography={topo2b.id}',
                 'select': f'/ui/api/selection/topography/{topo2b.id}/select/',
                 'analyze': topo2b_analyze,
                 'unselect': f'/ui/api/selection/topography/{topo2b.id}/unselect/'}
    }

    expected_dict_surface1 = {
        'category': 'exp',
        'category_name': 'Experimental data',
        'children': [expected_dict_topo1a, expected_dict_topo1b],
        'creator': user_url,
        'creator_name': user.name,
        'description': '',
        'folder': True,
        'key': f'surface-{surface1.pk}',
        'label': surface1.label,
        'name': surface1.name,
        'id': surface1.pk,
        'publication_authors': None,
        'publication_date': None,
        'publication_license': None,
        'publication_doi': None,
        'selected': False,
        'sharing_status': 'own',
        'tags': ['bike', 'train/tgv'],
        'title': surface1.name,
        'topography_count': 2,
        'type': 'surface',
        'version': None,
        'urls': {'analyze': surface1_analyze,
                 'download': f'/manager/api/surface/{surface1.id}/download/',
                 'detail': f'/ui/html/surface/?surface={surface1.id}',
                 'select': f'/ui/api/selection/surface/{surface1.id}/select/',
                 'unselect': f'/ui/api/selection/surface/{surface1.id}/unselect/'}
    }

    bike_pk = TagModel.objects.get(name='bike').pk
    train_pk = TagModel.objects.get(name='train').pk
    train_ice_pk = TagModel.objects.get(name='train/ice').pk
    train_tgv_pk = TagModel.objects.get(name='train/tgv').pk
    train_ice_restaurant_pk = TagModel.objects.get(name='train/ice/restaurant').pk

    bike_select_prefix = f"/ui/api/selection/tag/{bike_pk}/"
    train_select_prefix = f"/ui/api/selection/tag/{train_pk}/"
    train_ice_select_prefix = f"/ui/api/selection/tag/{train_ice_pk}/"
    train_tgv_select_prefix = f"/ui/api/selection/tag/{train_tgv_pk}/"
    train_ice_restaurant_select_prefix = f"/ui/api/selection/tag/{train_ice_restaurant_pk}/"

    expected_dicts = [
        {
            'title': 'bike',
            'type': 'tag',
            'version': None,
            'publication_authors': None,
            'publication_date': None,
            'id': bike_pk,
            'key': f'tag-{bike_pk}',
            'folder': True,
            'label': 'bike',
            'name': 'bike',
            'selected': False,
            'children': [
                # surface1, topo2a
                expected_dict_topo2a,
                expected_dict_surface1,
            ],
            'urls': {
                'select': bike_select_prefix + 'select/',
                'unselect': bike_select_prefix + 'unselect/'
            }
        },
        {
            'title': 'train',
            'type': 'tag',
            'version': None,
            'publication_authors': None,
            'publication_date': None,
            'id': train_pk,
            'key': f"tag-{train_pk}",
            'folder': True,
            'label': 'train',
            'name': 'train',
            'selected': False,
            'children': [
                {
                    'title': 'ice',
                    'type': 'tag',
                    'version': None,
                    'id': train_ice_pk,
                    'publication_authors': None,
                    'publication_date': None,
                    'key': f"tag-{train_ice_pk}",
                    'folder': True,
                    'label': 'ice',
                    'name': 'train/ice',
                    'selected': False,
                    'children': [
                        # topo2a
                        expected_dict_topo2a,
                        {
                            'title': 'restaurant',
                            'type': 'tag',
                            'version': None,
                            'id': train_ice_restaurant_pk,
                            'publication_authors': None,
                            'publication_date': None,
                            'key': f"tag-{train_ice_restaurant_pk}",
                            'folder': True,
                            'label': 'restaurant',
                            'name': 'train/ice/restaurant',
                            'selected': False,
                            'children': [
                                # topo2b
                                expected_dict_topo2b
                            ],
                            'urls': {
                                'select': train_ice_restaurant_select_prefix + 'select/',
                                'unselect': train_ice_restaurant_select_prefix + 'unselect/'
                            }
                        }
                    ],
                    'urls': {
                        'select': train_ice_select_prefix + 'select/',
                        'unselect': train_ice_select_prefix + 'unselect/'
                    }
                },
                {
                    'title': 'tgv',
                    'type': 'tag',
                    'version': None,
                    'id': train_tgv_pk,
                    'publication_authors': None,
                    'publication_date': None,
                    'key': f"tag-{train_tgv_pk}",
                    'folder': True,
                    'label': 'tgv',
                    'name': 'train/tgv',
                    'selected': False,
                    'children': [
                        # surface1
                        expected_dict_surface1
                    ],
                    'urls': {
                        'select': train_tgv_select_prefix + 'select/',
                        'unselect': train_tgv_select_prefix + 'unselect/'
                    }
                }
            ],
            'urls': {
                'select': train_select_prefix + 'select/',
                'unselect': train_select_prefix + 'unselect/'
            }
        },

    ]

    resulted_dicts = ordereddicts_to_dicts(response.data['page_results'], sorted_by='title')
    assert resulted_dicts == expected_dicts

    #
    # Now restrict result by query parameters, search for "topo2a"
    #
    request = factory.get(reverse('ce_ui:tag-list') + f"?search={topo2a.name}")
    request.user = user
    request.session = session

    #
    # Create tag tree and compare with expectation
    #
    response = TagTreeView.as_view()(request)

    assert response.status_code == 200

    # only tags "bike" and "train/ice" should be included
    # all other tags should be missing
    expected_dicts = [
        {
            'title': 'bike',
            'type': 'tag',
            'version': None,
            'id': bike_pk,
            'publication_authors': None,
            'publication_date': None,
            'key': f'tag-{bike_pk}',
            'folder': True,
            'label': 'bike',
            'name': 'bike',
            'selected': False,
            'children': [
                # only topo2a is matching
                expected_dict_topo2a,
            ],
            'urls': {
                'select': bike_select_prefix + 'select/',
                'unselect': bike_select_prefix + 'unselect/'
            }
        },
        {
            'title': 'train',
            'type': 'tag',
            'version': None,
            'id': train_pk,
            'publication_authors': None,
            'publication_date': None,
            'key': f"tag-{train_pk}",
            'folder': True,
            'label': 'train',
            'name': 'train',
            'selected': False,
            'children': [
                {
                    'title': 'ice',
                    'type': 'tag',
                    'version': None,
                    'id': train_ice_pk,
                    'publication_authors': None,
                    'publication_date': None,
                    'key': f"tag-{train_ice_pk}",
                    'folder': True,
                    'label': 'ice',
                    'name': 'train/ice',
                    'selected': False,
                    'children': [
                        # topo2a
                        expected_dict_topo2a
                    ],
                    'urls': {
                        'select': train_ice_select_prefix + 'select/',
                        'unselect': train_ice_select_prefix + 'unselect/'
                    }
                },

            ],
            'urls': {
                'select': train_select_prefix + 'select/',
                'unselect': train_select_prefix + 'unselect/'
            }
        },

    ]
    resulted_dicts = ordereddicts_to_dicts(response.data['page_results'], sorted_by='title')
    assert resulted_dicts == expected_dicts

    #
    # Now restrict result by query parameters, search for category 'dum'
    # -> no result, because surface 3 would match, but has no tag
    #
    request = factory.get(reverse('ce_ui:tag-list') + "?category=dum")
    request.user = user
    request.session = session

    #
    # Create tag tree and compare with expectation
    #
    response = TagTreeView.as_view()(request)

    assert response.status_code == 200

    # no results expected
    resulted_dicts = ordereddicts_to_dicts(response.data['page_results'], sorted_by='title')
    assert resulted_dicts == []

    #
    # Now create another surface and share with this active user, than filter only for shared
    #
    user2 = UserFactory()
    surface4 = SurfaceFactory(creator=user2)
    surface4.tags = ['shared']
    surface4.save()
    surface4.share(user)

    shared_pk = TagModel.objects.get(name='shared').pk
    shared_prefix = f"/ui/html/tag/{shared_pk}/"
    surface4_prefix = f"/ui/html/surface/{surface4.pk}/"

    request = factory.get(reverse('ce_ui:tag-list') + "?sharing_status=shared_ingress")
    request.user = user
    request.session = session

    #
    # Create tag tree and compare with expectation
    #
    response = TagTreeView.as_view()(request)

    assert response.status_code == 200

    expected_dicts = [
        {
            'title': 'shared',
            'type': 'tag',
            'version': None,
            'id': shared_pk,
            'publication_authors': None,
            'publication_date': None,
            'key': f"tag-{shared_pk}",
            'folder': True,
            'label': 'shared',
            'name': 'shared',
            'selected': False,
            'children': [
                {
                    'title': surface4.name,
                    'type': 'surface',
                    'version': None,
                    'id': surface4.pk,
                    'publication_authors': None,
                    'publication_date': None,
                    'key': f"surface-{surface4.pk}",
                    'folder': False,
                    'label': surface4.label,
                    'name': surface4.name,
                    'selected': False,
                    'children': [],
                    'urls': {
                        'select': surface4_prefix + 'select/',
                        'unselect': surface4_prefix + 'unselect/'
                    }
                },
            ],
            'urls': {
                'select': shared_prefix + 'select/',
                'unselect': shared_prefix + 'unselect/'
            }
        },
    ]


@pytest.mark.django_db
def test_search_expressions_with_request_factory():
    user = UserFactory()

    surface1 = SurfaceFactory(creator=user)

    topo1a = Topography1DFactory(surface=surface1, description="a big tiger")
    topo1b = Topography1DFactory(surface=surface1, description="a big elephant")
    topo1c = Topography1DFactory(surface=surface1, description="Find this here and a small ant")
    topo1d = Topography1DFactory(surface=surface1, description="not me, big snake")

    surface2 = SurfaceFactory(creator=user)

    topo2a = Topography1DFactory(surface=surface2, name='Measurement 2A')
    Topography1DFactory(surface=surface2, name='Measurement 2B', description="a small lion")

    #
    # Set some tags
    #
    topo1b.tags = ['bike']
    topo1b.save()
    topo1c.tags = ['transport/bike']
    topo1c.save()
    topo1d.tags = ['bike']
    topo1d.save()

    #
    # Define helper function for testing searching
    #
    factory = APIRequestFactory()

    # simple search for a topography by name given a phrase
    result = search_surfaces(factory, user, f"'{topo2a.name}'")
    assert len(result) == 1  # one surface
    assert len(result[0]['children']) == 1  # one topography
    assert result[0]['children'][0]['name'] == topo2a.name

    # AND search for topographies by name
    result = search_surfaces(factory, user, f'"{topo2a.name}" "{topo1a.name}"')
    assert len(result) == 0  # no surfaces

    # OR search for topographies by name
    result = search_surfaces(factory, user, f'"{topo2a.name}" OR "{topo1a.name}"')
    assert len(result) == 2  # two surfaces
    # noinspection DuplicatedCode
    assert len(result[0]['children']) == 1  # one topography
    assert len(result[1]['children']) == 1  # one topography
    assert result[0]['children'][0]['name'] == topo1a.name
    assert result[1]['children'][0]['name'] == topo2a.name

    # Exclusion using '-'
    result = search_surfaces(factory, user, '-elephant')
    assert len(result) == 2
    assert result[0]['name'] == surface1.name
    assert result[1]['name'] == surface2.name
    assert len(result[0]['children']) == 3  # here one measurement is excluded
    assert len(result[1]['children']) == 2

    # Searching nearby
    result = search_surfaces(factory, user, 'Find * here')
    assert len(result) == 1
    assert result[0]['name'] == surface1.name
    assert len(result[0]['children']) == 1  # here one measurement has it
    assert result[0]['children'][0]["description"] == "Find this here and a small ant"

    # more complex search expression using a phrase
    #
    # Parentheses do not work with 'websearch' for simplicity.
    #
    # (NOT) binds most tightly, "quoted text" (FOLLOWED BY) next most tightly,
    # then AND (default if no parameter), with OR binding the least tightly.
    #

    # result = search_surfaces(f'bike AND "a big" or "a small" -"not me"')
    result = search_surfaces(factory, user, 'bike -snake big')

    assert len(result) == 1  # surface 2 is excluded because there is no "bike"
    assert result[0]['name'] == surface1.name
    assert len(result[0]['children']) == 1
    assert result[0]['children'][0]["name"] == topo1b.name  # topo1d is excluded because of 'not me'


@pytest.mark.django_db
def test_search_for_user_with_request_factory():
    user1 = UserFactory(name="Bob Marley")
    user2 = UserFactory(name="Bob Dylan")

    surf1 = SurfaceFactory(creator=user1)
    surf2 = SurfaceFactory(creator=user2)

    request_factory = APIRequestFactory()

    #
    # So far nothing has been shared
    #
    # User 1 searches
    result = search_surfaces(request_factory, user1, "Bob")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 0

    result = search_surfaces(request_factory, user1, "Marley")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 0

    result = search_surfaces(request_factory, user1, "Dylan")
    assert len(result) == 0

    # User 2 searches
    result = search_surfaces(request_factory, user2, "Bob")
    assert len(result) == 1
    assert result[0]['name'] == surf2.name
    assert len(result[0]['children']) == 0

    result = search_surfaces(request_factory, user2, "Marley")
    assert len(result) == 0

    result = search_surfaces(request_factory, user2, "Dylan")
    assert len(result) == 1
    assert result[0]['name'] == surf2.name
    assert len(result[0]['children']) == 0

    #
    # User1 shares his surface with user2
    #
    surf1.set_permissions(user2, 'edit')

    # User 2 searches, now surface of user 1 is also visible
    result = search_surfaces(request_factory, user2, "Bob")
    assert len(result) == 2
    assert set(r['name'] for r in result) == set((surf1.name, surf2.name))
    assert len(result[0]['children']) == 0
    assert len(result[1]['children']) == 0

    result = search_surfaces(request_factory, user2, "Marley")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 0

    result = search_surfaces(request_factory, user2, "Dylan")
    assert len(result) == 1
    assert result[0]['name'] == surf2.name
    assert len(result[0]['children']) == 0

    #
    # User1 adds a topography to shared surface, it should be findable by both users using first user's name
    #
    topo1a = Topography1DFactory(surface=surf1, creator=user1)

    # User 1 searches, finds also topography
    result = search_surfaces(request_factory, user1, "Bob")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 1

    result = search_surfaces(request_factory, user1, "Marley")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 1

    result = search_surfaces(request_factory, user1, "Dylan")
    assert len(result) == 0

    # User 2 searches, finds also topography of user 1 in shared surface
    result = search_surfaces(request_factory, user2, "Bob")
    assert len(result) == 2
    assert set(r['name'] for r in result) == set((surf1.name, surf2.name))
    assert len(result[0]['children']) == 1
    assert len(result[1]['children']) == 0  # user2's own surface has no topography

    result = search_surfaces(request_factory, user2, "Marley")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 1

    result = search_surfaces(request_factory, user2, "Dylan")
    assert len(result) == 1
    assert result[0]['name'] == surf2.name
    assert len(result[0]['children']) == 0

    #
    # User2 adds a topography to shared surface, it should be findable by both users using user2's last name
    #
    topo1b = Topography1DFactory(surface=surf1, creator=user2)

    # User 1 searches, finds topographies, depending on search term
    result = search_surfaces(request_factory, user1, "Bob")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 2

    result = search_surfaces(request_factory, user1, "Marley")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 1  # topography uploaded by user2 should not be shown
    assert result[0]['children'][0]['name'] == topo1a.name

    result = search_surfaces(request_factory, user1, "Dylan")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name  # now own surface is also listed with one topography matching "Dylan"
    assert len(result[0]['children']) == 1  # topography uploaded by user2 should be shown alone
    assert result[0]['children'][0]['name'] == topo1b.name

    # User 2 searches, finds also topography of user 1 in shared surface
    result = search_surfaces(request_factory, user2, "Bob")
    assert len(result) == 2
    assert set(r['name'] for r in result) == set((surf1.name, surf2.name))
    assert len(result[0]['children']) == 2
    assert len(result[1]['children']) == 0  # user2's own surface has no topography

    result = search_surfaces(request_factory, user2, "Marley")
    assert len(result) == 1
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 1  # topography uploaded by user1 should be shown alone
    assert result[0]['children'][0]['name'] == topo1a.name

    result = search_surfaces(request_factory, user2, "Dylan")
    assert len(result) == 2
    assert set(r['name'] for r in result) == set((surf1.name, surf2.name))  # now also surf1 is listed
    assert result[0]['name'] == surf1.name
    assert len(result[0]['children']) == 1  # topography uploaded by user1 should be shown alone
    assert result[0]['children'][0]['name'] == topo1b.name
    assert result[1]['name'] == surf2.name
    assert len(result[1]['children']) == 0



"""Test related to Find&elect tab except of searching"""

import pytest

from django.shortcuts import reverse
from rest_framework.test import APIRequestFactory

from topobank.manager.tests.utils import SurfaceFactory, UserFactory, Topography1DFactory, TagModelFactory
from topobank.utils import assert_no_form_errors

from ..utils import selected_instances
from ..views import select_surface, unselect_surface, select_topography, unselect_topography, select_tag, \
    unselect_tag, unselect_all, DEFAULT_SELECT_TAB_STATE


@pytest.mark.django_db
def test_select_surface():
    user = UserFactory()
    surface1 = SurfaceFactory(creator=user)
    surface2 = SurfaceFactory(creator=user)
    surface3 = SurfaceFactory(creator=user)
    topo3a = Topography1DFactory(surface=surface3)
    Topography1DFactory(surface=surface3)

    factory = APIRequestFactory()
    session = {}

    #
    # First select a single surface
    #
    request = factory.post(reverse('ce_ui:surface-select', kwargs=dict(pk=surface1.pk)))
    request.user = user
    request.session = session

    response = select_surface(request, surface1.pk)

    assert response.status_code == 200

    assert request.session['selection'] == [f'surface-{surface1.pk}']

    assert selected_instances(request)[1] == [surface1]

    #
    # Then select another
    #
    request = factory.post(reverse('ce_ui:surface-select', kwargs=dict(pk=surface1.pk)))
    request.user = user
    request.session = session

    response = select_surface(request, surface2.pk)

    assert response.status_code == 200

    assert sorted(request.session['selection']) == [f'surface-{surface1.pk}', f'surface-{surface2.pk}']

    assert selected_instances(request)[1] == [surface1, surface2]

    #
    # If a surface is selected after a single topography of this surface
    # was selected, the selection of the single topography should be still there
    #
    request = factory.post(reverse('ce_ui:topography-select', kwargs=dict(pk=topo3a.pk)))
    request.user = user
    request.session = session

    response = select_topography(request, topo3a.pk)

    assert response.status_code == 200

    assert sorted(request.session['selection']) == [f'surface-{surface1.pk}', f'surface-{surface2.pk}',
                                                    f'topography-{topo3a.pk}']

    assert selected_instances(request)[0] == [topo3a]
    assert selected_instances(request)[1] == [surface1, surface2]

    request = factory.post(reverse('ce_ui:surface-select', kwargs=dict(pk=surface3.pk)))
    request.user = user
    request.session = session

    response = select_surface(request, surface3.pk)

    assert response.status_code == 200

    # the selection for the single topography should still be present
    assert sorted(request.session['selection']) == [f'surface-{surface1.pk}', f'surface-{surface2.pk}',
                                                    f'surface-{surface3.pk}', f'topography-{topo3a.pk}']

    assert selected_instances(request)[0] == [topo3a]
    assert selected_instances(request)[1] == [surface1, surface2, surface3]


@pytest.mark.django_db
def test_unselect_surface():
    user = UserFactory()
    surface1 = SurfaceFactory(creator=user)
    surface2 = SurfaceFactory(creator=user)

    factory = APIRequestFactory()
    session = dict(selection=[f'surface-{surface1.pk}', f'surface-{surface2.pk}'])

    #
    # deselect a surface
    #
    request = factory.post(reverse('ce_ui:surface-unselect', kwargs=dict(pk=surface1.pk)))
    request.user = user
    request.session = session

    response = unselect_surface(request, surface1.pk)

    assert response.status_code == 200

    assert request.session['selection'] == [f'surface-{surface2.pk}']

    assert selected_instances(request)[1] == [surface2]


@pytest.mark.django_db
def test_try_to_select_surface_but_not_allowed():
    user1 = UserFactory()
    user2 = UserFactory()
    surface1 = SurfaceFactory(creator=user1)

    factory = APIRequestFactory()
    session = {}

    request = factory.post(reverse('ce_ui:surface-select', kwargs=dict(pk=surface1.pk)))
    request.user = user2
    request.session = session

    response = select_surface(request, surface1.pk)

    assert response.status_code == 403


@pytest.mark.django_db(transaction=True)
def test_try_to_select_topography_but_not_allowed():
    user1 = UserFactory()
    user2 = UserFactory()
    surface1 = SurfaceFactory(creator=user1)
    topo1 = Topography1DFactory(surface=surface1)

    factory = APIRequestFactory()
    session = {}

    request = factory.post(reverse('ce_ui:topography-select', kwargs=dict(pk=topo1.pk)))
    request.user = user2
    request.session = session

    response = select_topography(request, topo1.pk)

    assert response.status_code == 403

    # if user 1 shares the surface with user 2, it is allowed
    surface1.share(user2)
    response = select_topography(request, topo1.pk)
    assert response.status_code == 200


@pytest.mark.django_db(transaction=True)
def test_try_to_select_tag_but_not_allowed():
    user1 = UserFactory()
    user2 = UserFactory()

    tag1 = TagModelFactory()
    SurfaceFactory(creator=user1, tags=[tag1])

    factory = APIRequestFactory()
    session = {}

    request = factory.post(reverse('ce_ui:tag-select', kwargs=dict(pk=tag1.pk)))
    request.user = user2
    request.session = session

    response = select_tag(request, tag1.pk)

    # not allowed, because tag is not used by user 2
    assert response.status_code == 403

    # If user 2 also uses this tag, it can be selected
    SurfaceFactory(creator=user2, tags=[tag1])

    response = select_tag(request, tag1.pk)
    assert response.status_code == 200


@pytest.mark.django_db
def test_select_topography():
    user = UserFactory()
    surface1 = SurfaceFactory(creator=user)
    topo1a = Topography1DFactory(surface=surface1)
    topo1b = Topography1DFactory(surface=surface1)
    topo1c = Topography1DFactory(surface=surface1)
    SurfaceFactory(creator=user)

    factory = APIRequestFactory()
    session = {}

    #
    # First select a single surface
    #
    request = factory.post(reverse('ce_ui:topography-select', kwargs=dict(pk=topo1a.pk)))
    request.user = user
    request.session = session

    response = select_topography(request, topo1a.pk)

    assert response.status_code == 200

    assert request.session['selection'] == [f'topography-{topo1a.pk}']

    assert selected_instances(request)[0] == [topo1a]

    #
    # Then select a second one
    #
    request = factory.post(reverse('ce_ui:topography-select', kwargs=dict(pk=topo1b.pk)))
    request.user = user
    request.session = session

    response = select_topography(request, topo1b.pk)

    assert response.status_code == 200

    assert sorted(request.session['selection']) == [f'topography-{topo1a.pk}', f'topography-{topo1b.pk}']

    assert selected_instances(request)[0] == [topo1a, topo1b]

    #
    # When selecting all topographies of a surface, the surface should not be selected
    #
    request = factory.post(reverse('ce_ui:topography-select', kwargs=dict(pk=topo1c.pk)))
    request.user = user
    request.session = session

    response = select_topography(request, topo1c.pk)

    assert response.status_code == 200

    assert sorted(request.session['selection']) == [f'topography-{topo1a.pk}', f'topography-{topo1b.pk}',
                                                    f'topography-{topo1c.pk}']

    assert selected_instances(request)[0] == [topo1a, topo1b, topo1c]

    assert selected_instances(request)[1] == []  # we only want explicitly selected objects now

    #
    # When selecting some arbitrary topography, a permission denied should show up
    #
    invalid_pk = 99999999999
    request = factory.post(reverse('ce_ui:topography-select', kwargs=dict(pk=invalid_pk)))
    request.user = user
    request.session = session

    response = select_topography(request, invalid_pk)

    assert response.status_code == 403


@pytest.mark.django_db
def test_unselect_topography():
    user = UserFactory()
    surface1 = SurfaceFactory(creator=user)
    topo1a = Topography1DFactory(surface=surface1)
    topo1b = Topography1DFactory(surface=surface1)
    surface2 = SurfaceFactory(creator=user)

    factory = APIRequestFactory()
    session = dict(selection=[f'surface-{surface1.pk}', f'surface-{surface2.pk}', f'topography-{topo1b.pk}'])

    #
    # deselect a topography
    #
    request = factory.post(reverse('ce_ui:topography-unselect', kwargs=dict(pk=topo1a.pk)))
    request.user = user
    request.session = session

    response = unselect_topography(request, topo1a.pk)

    assert response.status_code == 200

    # This has no effect, since the topography was not explicitly selected
    assert sorted(request.session['selection']) == sorted([f'surface-{surface1.pk}', f'surface-{surface2.pk}',
                                                           f'topography-{topo1b.pk}'])

    assert selected_instances(request)[0] == [topo1b]
    assert selected_instances(request)[1] == [surface1, surface2]

    #
    # Now remove topo1b
    #
    request = factory.post(reverse('ce_ui:topography-unselect', kwargs=dict(pk=topo1b.pk)))
    request.user = user
    request.session = session

    response = unselect_topography(request, topo1b.pk)

    assert response.status_code == 200
    assert sorted(request.session['selection']) == [f'surface-{surface1.pk}', f'surface-{surface2.pk}']
    assert selected_instances(request)[0] == []
    assert selected_instances(request)[1] == [surface1, surface2]


#
# Tests for selection of tags
#
@pytest.mark.django_db
def test_select_tag():
    user = UserFactory()

    tag1 = TagModelFactory()
    tag2 = TagModelFactory()

    # we use the tags, so the user is allowed to select it
    SurfaceFactory(creator=user, tags=[tag1, tag2])

    factory = APIRequestFactory()
    session = {}

    #
    # First select a single tag
    #
    request = factory.post(reverse('ce_ui:tag-select', kwargs=dict(pk=tag1.pk)))
    request.user = user
    request.session = session

    response = select_tag(request, tag1.pk)

    assert response.status_code == 200

    assert request.session['selection'] == [f'tag-{tag1.pk}']

    assert selected_instances(request)[2] == [tag1]

    #
    # Then select another
    #
    request = factory.post(reverse('ce_ui:tag-select', kwargs=dict(pk=tag2.pk)))
    request.user = user
    request.session = session

    response = select_tag(request, tag2.pk)

    assert response.status_code == 200

    assert sorted(request.session['selection']) == [f'tag-{tag1.pk}', f'tag-{tag2.pk}']

    assert selected_instances(request)[2] == [tag1, tag2]


@pytest.mark.django_db
def test_unselect_tag():
    user = UserFactory()

    tag1 = TagModelFactory()
    tag2 = TagModelFactory()

    # we use the tags, so the user is allowed to select it
    SurfaceFactory(creator=user, tags=[tag1, tag2])

    factory = APIRequestFactory()
    session = dict(selection=[f'tag-{tag1.pk}', f'tag-{tag2.pk}'])

    #
    # deselect a tag
    #
    request = factory.post(reverse('ce_ui:tag-unselect', kwargs=dict(pk=tag1.pk)))
    request.user = user
    request.session = session

    response = unselect_tag(request, tag1.pk)

    assert response.status_code == 200

    assert request.session['selection'] == [f'tag-{tag2.pk}']

    assert selected_instances(request)[2] == [tag2]


@pytest.mark.django_db
def test_unselect_all():
    user = UserFactory()

    tag1 = TagModelFactory()

    # we use the tags, so the user is allowed to select it
    surface1 = SurfaceFactory(creator=user, tags=[tag1])
    topo1 = Topography1DFactory(surface=surface1)

    factory = APIRequestFactory()
    session = dict(selection=[f'tag-{tag1.pk}', f'surface-{surface1.pk}', f'topography-{topo1.pk}'])

    #
    # deselect all
    #
    request = factory.post(reverse('ce_ui:unselect-all'))
    request.user = user
    request.session = session

    response = unselect_all(request)

    assert response.status_code == 200

    assert request.session['selection'] == []


@pytest.mark.django_db
def test_select_tab_state_should_be_default_after_login(client):
    # first request the site anonymously .. select tab state is set to that of
    # an anonymous user
    response = client.get(reverse('ce_ui:select'))
    assert response.context['select_tab_state']['sharing_status'] == 'published_ingress'

    # Then login as authenticated user
    password = "abcd"
    user = UserFactory(password=password)

    # we use a real request in order to trigger the signal
    response = client.post(reverse('account_login'), {
        'password': password,
        'login': user.username,
    })

    assert response.status_code == 302
    assert_no_form_errors(response)

    response = client.get(reverse('ce_ui:select'))

    assert response.context['select_tab_state'] == DEFAULT_SELECT_TAB_STATE


@pytest.mark.django_db
def test_select_tab_state_should_be_default_after_search(client, handle_usage_statistics):
    state_before_search = DEFAULT_SELECT_TAB_STATE.copy()
    state_before_search['current_page'] = 2

    user = UserFactory()
    client.force_login(user)
    client.session['select_tab_state'] = state_before_search

    response = client.get(reverse('ce_ui:select'), data={'search': 'what I want to find'})

    exp_state_after_search = DEFAULT_SELECT_TAB_STATE.copy()
    exp_state_after_search['search_term'] = 'what I want to find'

    assert exp_state_after_search['current_page'] == 1

    assert response.context['select_tab_state'] == exp_state_after_search



import logging

from django import shortcuts

from guardian.shortcuts import get_perms
from rest_framework import serializers
from tagulous.contrib.drf import TagRelatedManagerField

from topobank.manager.models import Surface, Topography, TagModel
from topobank.manager.serializers import SurfaceSerializer, TopographySerializer
from topobank.manager.utils import subjects_to_base64

from .utils import get_search_term, filtered_topographies

_log = logging.getLogger(__name__)


# From: RomanKhudobei, https://github.com/encode/django-rest-framework/issues/1655
class TopographySearchSerializer(serializers.ModelSerializer):
    class Meta:
        model = Topography
        fields = ['id', 'type', 'name', 'creator', 'description', 'tags', 'urls', 'selected', 'key', 'surface_key',
                  'title', 'folder', 'version', 'publication_date', 'publication_authors', 'datafile_format',
                  'measurement_date', 'resolution_x', 'resolution_y', 'size_x', 'size_y', 'size_editable', 'unit',
                  'unit_editable', 'height_scale', 'height_scale_editable', 'creator_name', 'sharing_status', 'label',
                  'is_periodic', 'thumbnail', 'tags', 'instrument_name', 'instrument_type', 'instrument_parameters']

    creator = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name='users:detail',
        lookup_field='username',
        default=serializers.CurrentUserDefault()
    )

    title = serializers.CharField(source='name', read_only=True)  # set this through name

    urls = serializers.SerializerMethodField()
    selected = serializers.SerializerMethodField()
    key = serializers.SerializerMethodField()
    surface_key = serializers.SerializerMethodField()
    sharing_status = serializers.SerializerMethodField()
    # `folder` is Fancytree-specific, see
    # https://wwwendt.de/tech/fancytree/doc/jsdoc/global.html#NodeData
    folder = serializers.BooleanField(default=False, read_only=True)
    tags = TagRelatedManagerField()
    # `type` should be the output of mangle_content_type(Meta.model)
    type = serializers.CharField(default='topography', read_only=True)
    version = serializers.CharField(default=None, read_only=True)
    publication_authors = serializers.CharField(default=None, read_only=True)
    publication_date = serializers.CharField(default=None, read_only=True)
    creator_name = serializers.SerializerMethodField()
    label = serializers.SerializerMethodField()

    def get_urls(self, obj):
        """Return only those urls which are usable for the user

        :param obj: topography object
        :return: dict with { url_name: url }
        """
        user = self.context['request'].user
        surface = obj.surface

        perms = get_perms(user, surface)  # TODO are permissions needed here?

        urls = {
            'select': shortcuts.reverse('ce_ui:topography-select', kwargs=dict(pk=obj.pk)),
            'unselect': shortcuts.reverse('ce_ui:topography-unselect', kwargs=dict(pk=obj.pk))
        }

        if 'view_surface' in perms:
            urls['detail'] = f"{shortcuts.reverse('ce_ui:topography-detail')}?topography={obj.pk}"
            urls['analyze'] = f"{shortcuts.reverse('ce_ui:results-list')}?subjects={subjects_to_base64([obj])}"

        return urls

    def get_selected(self, obj):
        try:
            topographies, surfaces, tags = self.context['selected_instances']
            return obj in topographies
        except KeyError:
            return False

    def get_key(self, obj):
        return f"topography-{obj.pk}"

    def get_surface_key(self, obj):
        return f"surface-{obj.surface.pk}"

    def get_sharing_status(self, obj):
        user = self.context['request'].user
        if hasattr(obj.surface, 'is_published') and obj.surface.is_published:
            return 'published'
        elif user == obj.surface.creator:
            return "own"
        else:
            return "shared"

    def get_creator_name(self, obj):
        return obj.creator.name

    def get_label(self, obj):
        return obj.label


class SurfaceSearchSerializer(serializers.ModelSerializer):
    class Meta:
        model = Surface
        fields = ['id', 'type', 'name', 'creator', 'creator_name', 'description', 'category', 'category_name', 'tags',
                  'children', 'sharing_status', 'urls', 'selected', 'key', 'title', 'folder', 'version',
                  'publication_doi', 'publication_date', 'publication_authors', 'publication_license',
                  'topography_count', 'label']

    creator = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name='users:detail',
        lookup_field='username',
        default=serializers.CurrentUserDefault()
    )

    title = serializers.CharField(source='name')
    children = serializers.SerializerMethodField()

    urls = serializers.SerializerMethodField()
    selected = serializers.SerializerMethodField()
    key = serializers.SerializerMethodField()
    # `folder` is Fancytree-specific, see
    # https://wwwendt.de/tech/fancytree/doc/jsdoc/global.html#NodeData
    folder = serializers.BooleanField(default=True, read_only=True)
    sharing_status = serializers.SerializerMethodField()
    tags = TagRelatedManagerField()
    # `type` should be the output of mangle_content_type(Meta.model)
    type = serializers.CharField(default='surface', read_only=True)
    version = serializers.SerializerMethodField()
    publication_date = serializers.SerializerMethodField()
    publication_authors = serializers.SerializerMethodField()
    publication_license = serializers.SerializerMethodField()
    publication_doi = serializers.SerializerMethodField()
    topography_count = serializers.SerializerMethodField()
    category_name = serializers.SerializerMethodField()
    creator_name = serializers.SerializerMethodField()
    label = serializers.SerializerMethodField()

    def get_children(self, obj):
        """Get serialized topographies for given surface.

        Parameters
        ----------
        obj : Surface

        Returns
        -------

        """
        #
        # We only want topographies as children which match the given search term,
        # if no search term is given, all topographies should be included
        #
        request = self.context['request']
        search_term = get_search_term(request)
        search_term_given = len(search_term) > 0

        # only filter topographies by search term if surface does not match search term
        # otherwise list all topographies
        if search_term_given:
            topographies = filtered_topographies(request, [obj])
        else:
            topographies = obj.topography_set.all()
        return TopographySearchSerializer(topographies, many=True, context=self.context).data

    def get_urls(self, obj):

        user = self.context['request'].user
        perms = get_perms(user, obj)  # TODO are permissions needed here?

        urls = {
            'select': shortcuts.reverse('ce_ui:surface-select', kwargs=dict(pk=obj.pk)),
            'unselect': shortcuts.reverse('ce_ui:surface-unselect', kwargs=dict(pk=obj.pk))
        }
        if 'view_surface' in perms:
            urls['detail'] = f"{shortcuts.reverse('ce_ui:surface-detail')}?surface={obj.pk}"
            if obj.num_topographies() > 0:
                urls.update({
                    'analyze': f"{shortcuts.reverse('ce_ui:results-list')}?subjects={subjects_to_base64([obj])}"
                })
            urls['download'] = shortcuts.reverse('manager:surface-download', kwargs=dict(surface_id=obj.id))

        return urls

    def get_selected(self, obj):
        try:
            topographies, surfaces, tags = self.context['selected_instances']
            return obj in surfaces
        except KeyError:
            return False

    def get_key(self, obj):
        return f"surface-{obj.pk}"

    def get_sharing_status(self, obj):
        user = self.context['request'].user
        if hasattr(obj, 'is_published') and obj.is_published:
            return 'published'
        elif user == obj.creator:
            return "own"
        else:
            return "shared"

    def get_version(self, obj):
        return obj.publication.version if obj.is_published else None

    def get_publication_date(self, obj):
        return obj.publication.datetime.date() if obj.is_published else None

    def get_publication_authors(self, obj):
        return obj.publication.get_authors_string() if obj.is_published else None

    def get_publication_license(self, obj):
        return obj.publication.license if obj.is_published else None

    def get_publication_doi(self, obj):
        return obj.publication.doi_url if obj.is_published else None

    def get_topography_count(self, obj):
        return obj.topography_set.count()

    def get_category_name(self, obj):
        return obj.get_category_display()

    def get_creator_name(self, obj):
        return obj.creator.name

    def get_label(self, obj):
        return obj.label


class TagSearchSerizalizer(serializers.ModelSerializer):
    class Meta:
        model = TagModel
        fields = ['id', 'key', 'type', 'title', 'name', 'children', 'folder', 'urls', 'selected', 'version',
                  'publication_date', 'publication_authors', 'label']

    children = serializers.SerializerMethodField()
    # `folder` is Fancytree-specific, see
    # https://wwwendt.de/tech/fancytree/doc/jsdoc/global.html#NodeData
    folder = serializers.BooleanField(default=True, read_only=True)
    key = serializers.SerializerMethodField()
    label = serializers.SerializerMethodField()
    publication_authors = serializers.CharField(default=None, read_only=True)
    publication_date = serializers.CharField(default=None, read_only=True)
    selected = serializers.SerializerMethodField()
    title = serializers.CharField(source='label', read_only=True)
    # `type` should be the output of mangle_content_type(Meta.model)
    type = serializers.CharField(default='tag', read_only=True)
    urls = serializers.SerializerMethodField()
    version = serializers.CharField(default=None, read_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._surface_serializer = SurfaceSerializer(context=self.context)
        self._topography_serializer = TopographySerializer(context=self.context)

    def get_urls(self, obj):
        urls = {
            'select': shortcuts.reverse('ce_ui:tag-select', kwargs=dict(pk=obj.pk)),
            'unselect': shortcuts.reverse('ce_ui:tag-unselect', kwargs=dict(pk=obj.pk))
        }
        return urls

    def get_key(self, obj):
        return f"tag-{obj.pk}"

    def get_selected(self, obj):
        topographies, surfaces, tags = self.context['selected_instances']
        return obj in tags

    def get_children(self, obj):
        result = []

        #
        # Assume that all surfaces and topographies given in the context are already filtered
        #
        surfaces = self.context['surfaces'].filter(tags__pk=obj.pk)  # .order_by('name')
        topographies = self.context['topographies'].filter(tags__pk=obj.pk)  # .order_by('name')
        tags = [x for x in obj.children.all() if x in self.context['tags_for_user']]

        #
        # Serialize children and append to this tag
        #
        result.extend(TopographySearchSerializer(topographies, many=True, context=self.context).data)
        result.extend(SurfaceSearchSerializer(surfaces, many=True, context=self.context).data)
        result.extend(TagSearchSerizalizer(tags, many=True, context=self.context).data)

        return result

    def get_label(self, obj):
        return obj.label



import pytest

from topobank.fixtures import handle_usage_statistics, sync_analysis_functions, test_analysis_function  # noqa: F401
from topobank.manager.tests.utils import two_topos, user_three_topographies_three_surfaces_three_tags, \
    UserFactory  # noqa: F401
from topobank.organizations.tests.utils import OrganizationFactory


@pytest.mark.django_db
@pytest.fixture
def user_with_plugin():
    org_name = "Test Organization"
    org = OrganizationFactory(name=org_name, plugins_available="ce_ui")
    user = UserFactory()
    user.groups.add(org.group)
    return user



import pytest
from django.urls import reverse
from django.contrib.auth.models import Permission

from topobank.users.tests.test_utils import are_collaborating, SurfaceFactory


@pytest.mark.django_db
def test_sharing_profile(client, django_user_model, handle_usage_statistics):
    user1 = django_user_model.objects.create_user(username='testuser1', password="abcd$1234")
    user2 = django_user_model.objects.create_user(username='testuser2', password="abcd$5678")

    # give both user permissions to skip all terms, we want to test independently from this
    skip_perm = Permission.objects.get(codename='can_skip_terms')
    user1.user_permissions.add(skip_perm)
    user2.user_permissions.add(skip_perm)

    assert client.login(username='testuser1', password='abcd$1234')

    response = client.get(reverse('users:detail', kwargs={'username': 'testuser1'}))  # same user
    assert response.status_code == 200

    #
    # both users don't share anything, so they can't see each others profiles
    #
    assert not are_collaborating(user1, user2)

    response = client.get(reverse('users:detail', kwargs={'username': 'testuser2'}))  # other user!!
    assert response.status_code == 403  # Forbidden

    # share sth. and the view of profile is possible
    surface = SurfaceFactory(creator=user1)
    surface.share(user2)

    assert are_collaborating(user1, user2)
    response = client.get(reverse('users:detail', kwargs={'username': 'testuser2'}))
    assert response.status_code == 200  # Allowed



"""
Template tags for displaying icons.
"""
from django import template

register = template.Library()


@register.inclusion_tag('manager/fa5_icon.html')
def fa5_icon(name, style_prefix='fas', title=None):
    """Returns a HMTL snippet which generates an fontawesome 5 icon.

    Parameters:

        style_prefix: str
            'fas' (default) for solid icons, 'far' for regular icons
    """
    return {
        'classes': f'fa-{name} {style_prefix}',
        'title': title,
    }



import logging

from django.dispatch import receiver

from allauth.account.signals import user_logged_in

from .views import DEFAULT_SELECT_TAB_STATE

_log = logging.getLogger(__name__)


@receiver(user_logged_in)
def set_default_select_tab_state(request, user, **kwargs):
    """At each login, the state of the select tab should be reset.
    """
    request.session['select_tab_state'] = DEFAULT_SELECT_TAB_STATE



import pytest

from django.test import SimpleTestCase
from django.conf import settings

from topobank.manager.tests.utils import UserFactory


@pytest.mark.skipif(settings.CHALLENGE_REDIRECT_URL == '', reason="No URL for challenge given in settings.")
@pytest.mark.django_db
def test_redirect_for_challenge(client):
    user = UserFactory()
    client.force_login(user)
    response = client.get('/challenge/')
    SimpleTestCase().assertRedirects(response, settings.CHALLENGE_REDIRECT_URL,
                                     status_code=302, fetch_redirect_response=False)



import logging

from django.db.models import Count, Q, Value, TextField
from django.db.models.functions import Replace
from django.core.exceptions import PermissionDenied
from django.contrib.postgres.search import SearchVector, SearchQuery

from rest_framework.reverse import reverse

from guardian.shortcuts import get_users_with_perms

from topobank.manager.utils import subjects_to_dict, surfaces_for_user
from topobank.manager.models import SurfaceCollection, TagModel, Topography, Surface

_log = logging.getLogger(__name__)

DEFAULT_DATASOURCE_NAME = 'Default'
MAX_LEN_SEARCH_TERM = 200
SELECTION_SESSION_VARNAME = 'selection'


def surface_collection_name(surface_names, max_total_length=SurfaceCollection.MAX_LENGTH_NAME):
    """For a given list of names, return a length-limited collection name."""
    num_surfaces = len(surface_names)
    k = 0
    coll_name_prefix = ""
    last_coll_name = ""
    while k < num_surfaces:
        coll_name_prefix += f"Surface '{surface_names[k]}'"
        num_rest = num_surfaces - (k + 1)
        coll_name = coll_name_prefix[:]
        if num_rest > 0:
            coll_name += f" and {num_rest} more"
        if len(coll_name) > max_total_length:
            if last_coll_name == "":
                coll_name = coll_name_prefix[:max_total_length - 4] + "..."
            else:
                coll_name = last_coll_name
            break
        else:
            last_coll_name = coll_name
            coll_name_prefix += ", "
            k += 1  # add one more and try if it still fits

    return coll_name


def selection_from_session(session):
    """Get selection from session.

    The selection is a list of strings like

    [ 'topography-1', 'surface-4']

    which represents the selected objects.
    """
    return session.get(SELECTION_SESSION_VARNAME, [])


def instances_to_selection(topographies=[], surfaces=[], tags=[]):
    """Returns a list of strings suitable for selecting instances.

    :param topographies: sequence of Topography instances
    :param surfaces: sequence of Surface instances
    :param tags: sequence of TagModel instances
    :return: list of str, alphabetically sorted
    """
    selection = []
    for topo in topographies:
        if topo.surface in surfaces:
            # will be included by surface
            continue
        selection.append(f'topography-{topo.id}')

    for surf in surfaces:
        selection.append(f'surface-{surf.id}')

    for tag in tags:
        selection.append(f'tag-{tag.id}')

    return sorted(selection)


def instances_to_topographies(topographies, surfaces, tags):
    """Returns a queryset of topographies, based on given instances

    Given topographies, surfaces and tags are resolved and
    all topographies are returned which are either
    - explicitly given
    - given indirectly by a surface
    - given indirectly by a tag, if the topography is tagged accordingly
    - given indirectly by a tag, if its surface is tagged accordingly

    Parameters
    ----------
    topographies: sequence of topographies
    surfaces: sequence of surfaces
    tags: sequence of tags

    Returns
    -------
    Queryset of topography, distinct
    """
    topography_ids = [topo.id for topo in topographies]
    surface_ids = [s.id for s in surfaces]
    tag_ids = [tag.id for tag in tags]

    topographies = Topography.objects.filter(id__in=topography_ids)
    topographies |= Topography.objects.filter(surface__in=surface_ids)
    topographies |= Topography.objects.filter(surface__tags__in=tag_ids)
    topographies |= Topography.objects.filter(tags__in=tag_ids)

    return topographies.distinct().order_by('id')


def instances_to_surfaces(surfaces, tags):
    """Returns a queryset of surfaces, based on given instances

    Given surfaces and tags are resolved and
    all surfaces are returned which are either
    - explicitly given
    - given indirectly by a tag, if the surface is tagged accordingly

    Parameters
    ----------
    surfaces: sequence of surfaces
    tags: sequence of tags

    Returns
    -------
    Queryset of surface, distinct
    """
    surface_ids = [s.id for s in surfaces]
    tag_ids = [tag.id for tag in tags]

    surfaces = Surface.objects.filter(id__in=surface_ids)
    surfaces |= Surface.objects.filter(tags__in=tag_ids)

    return surfaces.distinct().order_by('id')


def selection_to_instances(selection):
    """Returns a tuple with querysets of explicitly selected topographies, surfaces, and tags.

    View permission is not checked.

    :param selection: selection list as saved in session
    :return: tuple (topographies, surfaces, tags)

    The returned tuple has 3 elements:

     'topographies': all topographies explicitly found in the selection
     'surfaces': all surfaces explicitly found in the selection (not only because its topography was selected)
     'tags': all tags explicitly found in the selection (not only because all related items are selected)

    Also surfaces without topographies are returned in 'surfaces' if selected.
    """
    topography_ids = set()
    surface_ids = set()
    tag_ids = set()

    for type_id in selection:
        type, id = type_id.split('-')
        id = int(id)
        if type == 'topography':
            topography_ids.add(id)
        elif type == 'surface':
            surface_ids.add(id)
        elif type == 'tag':
            tag_ids.add(id)

    topographies = Topography.objects.filter(id__in=topography_ids)
    surfaces = Surface.objects.filter(id__in=surface_ids)
    tags = TagModel.objects.filter(id__in=tag_ids)

    return topographies, surfaces, tags


def selected_instances(request):
    """Return a tuple with topography, surface, and tag instances which are currently selected.

    View permission is checked for the user of the request.

    :request: HTTP request
    :return: tuple (topographies, surfaces, tags)

    The returned tuple has 3 elements, each a list:

     'topographies': all topographies in the selection
     'surfaces': all surfaces explicitly found in the selection (not only because its topography was selected)
     'tags': all tags explicitly found in selection (not because all tagged items are selected)

    Also surfaces without topographies are returned in 'surfaces' if selected.

    If only one topography is selected, it's surface is *not* returned in 'surfaces'.
    If a surface is explicitly selected, all of its topographies are contained in 'topographies'.
    """
    selection = selection_from_session(request.session)
    topographies, surfaces, tags = selection_to_instances(selection)

    # make sure that only topographies with read permission can be found here
    unique_surfaces = set(t.surface for t in topographies) | set(surfaces)
    surfaces_with_view_permission = [s for s in unique_surfaces if request.user.has_perm('view_surface', s)]
    topographies = [t for t in topographies if t.surface in surfaces_with_view_permission]
    surfaces = [s for s in surfaces if s in surfaces_with_view_permission]

    return topographies, surfaces, list(tags)


def current_selection_as_surface_list(request):
    """Returns a list of surfaces related to the current selection.

    For all selected items, surfaces, topographies, or tags
    the surface is identified which contains the selected data.
    In the result, each of those surfaces is included once.

    :param request: current request
    :return: list of Surface instances, sorted by name
    """
    topographies, surfaces, tags = selected_instances(request)

    #
    # Collect all surfaces related to the selected items in a set
    #
    surfaces = set(surfaces)
    for topo in topographies:
        surfaces.add(topo.surface)
    for tag in tags:
        related_objects = tag.get_related_objects(flat=True)
        for obj in related_objects:
            if isinstance(obj, Surface):
                surfaces.add(obj)
            elif hasattr(obj, 'surface'):
                surfaces.add(obj.surface)
    #
    # Filter surfaces such that the requesting user has permissions to read
    #
    surfaces = [surf for surf in surfaces if request.user.has_perm('view_surface', surf)]
    surfaces.sort(key=lambda s: s.name)

    return surfaces


def instances_to_basket_items(topographies, surfaces, tags):
    """

    Parameters
    ----------
    topographies
    surfaces
    tags

    Returns
    -------
    List of items in the basket. Each is a dict with keys

     label, type, unselect_url, key

    Example with one selected surface:

     [ {'label': "Test Surface",
        'type': "surface",
        'unselect_url': ".../manager/surface/13/unselect",
        'key': "surface-13"}
     ]

    """
    basket_items = []
    for surface in surfaces:
        unselect_url = reverse('ce_ui:surface-unselect', kwargs=dict(pk=surface.pk))
        basket_items.append(dict(label=str(surface),
                                 type="surface",
                                 id=surface.pk,
                                 unselect_url=unselect_url,
                                 key=f"surface-{surface.pk}"))
    for topography in topographies:
        unselect_url = reverse('ce_ui:topography-unselect', kwargs=dict(pk=topography.pk))
        basket_items.append(dict(label=topography.name,
                                 type="topography",
                                 id=topography.pk,
                                 unselect_url=unselect_url,
                                 key=f"topography-{topography.pk}",
                                 surface_key=f"surface-{topography.surface.pk}"))
    for tag in tags:
        unselect_url = reverse('ce_ui:tag-unselect', kwargs=dict(pk=tag.pk))
        basket_items.append(dict(label=tag.name,
                                 type="tag",
                                 id=tag.pk,
                                 unselect_url=unselect_url,
                                 key=f"tag-{tag.pk}"))

    return basket_items


def current_selection_as_basket_items(request):
    """Returns current selection as JSON suitable for the basket.

    Parameters
    ----------
    request

    Returns
    -------
    List of items in the basket. Each is a dict with keys

     label, type, unselect_url, key

    Example with one selected surface:

     [ {'label': "Test Surface",
        'type': "surface",
        'unselect_url': ".../manager/surface/13/unselect",
        'key': "surface-13"}
     ]

    """
    topographies, surfaces, tags = selected_instances(request)
    return instances_to_basket_items(topographies, surfaces, tags)


def filter_queryset_by_search_term(qs, search_term, search_fields):
    """Filter queryset for a given search term.

    Parameters
    ----------
    qs : QuerySet
        QuerySet which should be additionally filtered by a search term.
    search_term: str
        Search term entered by the user. Can be an expression.
        See https://docs.djangoproject.com/en/3.2/ref/contrib/postgres/search/
        for details.
    search_fields: list of str
        ORM expressions which refer to search fields, e.g. "description"
        or "topography__description" for the description field of a child object

    Returns
    -------
    Filtered query set.
    """
    return qs.annotate(
        search=SearchVector(*search_fields, config='english')
    ).filter(
        search=SearchQuery(search_term, config="english", search_type='websearch')
        # search__icontains=search_term  # alternative, which finds substrings but does not allow for expressions
    ).distinct('id').order_by('id')


def filtered_topographies(request, surfaces):
    """Return topographies which match a request.

    Parameters
    ----------
    request
    surfaces
        queryset with surfaces which already match

    Returns
    -------
    queryset with matching topographies

    """
    topographies = Topography.objects.filter(surface__in=surfaces)
    search_term = get_search_term(request)
    if search_term:
        # We introduce an extra field for search in tag names where the tag names
        # are changed so that the tokenizer splits the names into multiple words
        topographies = topographies.annotate(
            tag_names_for_search=Replace(
                Replace('tags__name', Value('.'), Value(' ')),  # replace . with space
                Value('/'), Value(' ')),  # replace / with space
            name_for_search=Replace('name', Value('.'), Value(' '), output_field=TextField())
        ).distinct('id').order_by('id')
        topographies = filter_queryset_by_search_term(topographies, search_term, [
            'description', 'creator__name', 'name_for_search', 'tag_names_for_search',
        ])
    return topographies


def tags_for_user(user, surfaces=None, topographies=None):
    """Return set of tags which can be used for autocomplete when editing tags.

    A user should not see all tags ever used on the app, but only those
    which were chosen by herself or collaborators and corresponding parent tags.

    :param user: User instance
    :param surfaces: surfaces to use, if None,
                     will be computed for given user; specify this to reuse previous calculation
                     or to reduce number of surfaces based on a request
    :param topographies: topographies to use, if None,
                     will be computed from surfaces; specify this to reuse previous calculation
                     or to reduce number of topographies based on a request
    :return: list of strings
    """
    if surfaces is None:
        surfaces = surfaces_for_user(user)
    if topographies is None:
        topographies = Topography.objects.filter(surface__in=surfaces)

    tags = TagModel.objects.filter(Q(surface__in=surfaces) | Q(topography__in=topographies))

    # add parent tags not already included
    for t in tags:
        tags |= t.get_ancestors()

    return tags.distinct()


def selection_to_subjects_dict(request):
    """Convert current selection into list of subjects as json.

    If 2 or more surfaces are created, also adds a SurfaceCollection
    instance to the subjects.

    Parameters
    ----------
    request

    Returns
    -------
    (eff_topographies, eff_surfaces, subjects_ids_json)

    where:
        eff_topographies       - list of topographies which are effectively
                                 included in the selection (selected explicitly
                                 or implicitly by surfaces and tags)
        eff_surfaces           - list of surfaces which are effectively
                                  included in the selection (by selecting surfaces+tags)
        subjects_ids_json - JSONfied dict with key: content type id, value: list of object ids
                            This dict encodes all subjects in the selection

    The eff_ results can be used for selection which tabs should be shown in the UI.
    This was the original purpose for returning them here.
    """
    topographies, surfaces, tags = selected_instances(request)
    effective_topographies = instances_to_topographies(topographies, surfaces, tags)
    effective_surfaces = instances_to_surfaces(surfaces, tags)

    # Do we have permission for all of these?
    user = request.user
    unique_surfaces = set(t.surface for t in effective_topographies) | set(effective_surfaces)
    surfaces_with_view_permission = [s for s in unique_surfaces if user.has_perm('view_surface', s)]
    effective_topographies = [t for t in effective_topographies if t.surface in surfaces_with_view_permission]
    effective_surfaces = [s for s in effective_surfaces if s in surfaces_with_view_permission]

    if len(effective_surfaces) > 1:
        # In order to find a matching SurfaceCollection, we need to search first
        # for all surface collections with same number of surfaces, then filtering
        # for the exact surfaces
        # (see https://stackoverflow.com/questions/16324362/django-queryset-get-exact-manytomany-lookup)
        surf_collections = SurfaceCollection.objects.annotate(surface_count=Count('surfaces')) \
            .filter(surface_count=len(effective_surfaces))
        for s in effective_surfaces:
            surf_collections = surf_collections.filter(surfaces__pk=s.pk)

        if surf_collections.count() > 0:  # should be exactly 0 or 1 but let's keep it robust here
            _log.info(f"Found existing surface collection for surfaces {[s.id for s in effective_surfaces]}.")
            coll = surf_collections.first()
            if surf_collections.count() > 1:
                _log.warning("More than on surface collection instance for surfaces "
                             f"{[s.id for s in effective_surfaces]} found.")
        else:
            coll = SurfaceCollection.objects.create(name=surface_collection_name([s.name for s in effective_surfaces]))
            coll.surfaces.set(effective_surfaces)
            coll.save()
            _log.info(f"Created new surface collection for surfaces {[s.id for s in effective_surfaces]}.")

        effective_surface_collections = [coll]
    else:
        effective_surface_collections = []

    # we collect effective topographies and surfaces because we have so far implementations
    # for analysis functions for topographies and surfaces
    subjects_ids = subjects_to_dict(effective_topographies + effective_surfaces + effective_surface_collections)

    return effective_topographies, effective_surfaces, subjects_ids


def get_search_term(request) -> str:
    """Extract a search term from given request.

    The search term is truncated at a maximum
    size of MAX_LEN_SEARCH_TERM.

    Parameters
    ----------
    request

    Returns
    -------
    String with search term, an empty string if no term was given.

    """
    print('GET', request.GET)
    search_term = request.GET.get('search', default='')
    search_term = search_term[:MAX_LEN_SEARCH_TERM]
    return search_term.strip()


def get_category(request) -> str:
    """Extract a surface category from given request.

    Parameters
    ----------
    request

    Returns
    -------
    String with requested category.

    Raises
    ------
    PermissionDenied() if an unknown category was given.
    """
    from .views import CATEGORY_FILTER_CHOICES
    category = request.GET.get('category', default='all')
    if category not in CATEGORY_FILTER_CHOICES.keys():
        raise PermissionDenied()
    return category


def get_sharing_status(request) -> str:
    """Extract a sharing status from given request.

     Parameters
     ----------
     request

     Returns
     -------
     String with requested sharing status.

     Raises
     ------
     PermissionDenied() if an unknown sharing status was given.
     """
    from .views import SHARING_STATUS_FILTER_CHOICES
    sharing_status = request.GET.get('sharing_status', default='all')
    if sharing_status not in SHARING_STATUS_FILTER_CHOICES.keys():
        raise PermissionDenied()
    return sharing_status


def get_tree_mode(request) -> str:
    """Extract tree_mode from given request.

     Parameters
     ----------
     request

     Returns
     -------
     String with requested tree mode.

     Raises
     ------
     PermissionDenied() if an unknown sharing status was given.
     """
    from .views import TREE_MODE_CHOICES
    tree_mode = request.GET.get('tree_mode', default='surface list')
    if tree_mode not in TREE_MODE_CHOICES:
        raise PermissionDenied()
    return tree_mode


def get_permission_table_data(instance, request_user, actions=['view', 'change', 'delete', 'share']):
    """Prepare data for a permission table.

    Parameters
    ----------
    instance: Model
        Model instance for which the permission table should be generated.
    request_user: User
        User for which the permission table is prepared.
    actions: list of str
        Action for which permission may be given. Example: ['view', 'change', 'delete', 'share']

    Returns
    -------
    List of tuples for cells. Each cell is represented by a 4-tuple

       (user_display_name, user_url, allowed, description)

    where

        user_display_name: name of the user
        user_url: URL to user details
        allowed: boolean which is true, when allowed
        description: a str with verbal description of this entry for tooltips.

    """
    perms = get_users_with_perms(instance, attach_perms=True)
    # is now a dict of the form
    #  <User: joe>: ['view_surface'], <User: dan>: ['view_surface', 'change_surface']}
    allowed_users = sorted(perms.keys(), key=lambda u: u.name if u else '')

    # convert to list of boolean based on list ACTIONS
    #
    # Each table element here is a 2-tuple: (cell content, cell title)
    #
    # The cell content is inserted into the cell.
    # The cell title is shown in a tooltip and can be used in tests.
    #
    perm_postfix = type(instance).__name__.lower()
    perms_table = []
    for user in allowed_users:

        if user == request_user:
            user_display_name = "You"
            auxiliary = "have"
        else:
            user_display_name = user.name
            auxiliary = "has"

        # the current user is represented as None, can be displayed in a special way in template ("You")
        row = [(user_display_name, user.get_absolute_url())]  # cell title is used for passing a link here
        for act in actions:

            perm = act + '_' + perm_postfix
            has_perm = perm in perms[user]

            cell_title = "{} {}".format(user_display_name, auxiliary)
            if not has_perm:
                cell_title += "n't"
            cell_title += " the permission to {} this surface".format(act)

            row.append((has_perm, cell_title))

        perms_table.append(row)

    return perms_table



from django import template

register = template.Library()

TABS_CONTEXT_KEYS = ['fixed_tabs', 'extra_tabs']


@register.inclusion_tag('tabnav/tabs.html', takes_context=True)
def tab_navigation(context):
    """Can be used to insert tab navigation on the page.

    Use {% tab_navigation %} in order to include
    navigation by tabs. This function here provides the
    context which is used with the given template.
    """

    if 'request' not in context:
        return {}  # Needed in case of 500 status (server error)

    is_anonymous = context['request'].user.is_anonymous
    request_path = context['request'].path
    active_tab = {}
    tabs = []
    for k in TABS_CONTEXT_KEYS:
        if k in context:
            for tab in context[k]:
                # make sure that all tabs have necessary keys
                tab.setdefault('login_required', True)

                # if user is not authenticated and login is required for
                # this tab, the tab can be excluded
                if tab['login_required'] and is_anonymous:
                    continue  # tab is skipped

                tab.setdefault('href', request_path)
                tab.setdefault('active', tab['href'].startswith(request_path))  # in order to make tab definition short
                tab.setdefault('icon', 'cog')  # just some icon that we see one is missing
                tab.setdefault('title', '')
                tab.setdefault('show_basket', False)
                tabs.append(tab)

                # we want to be able to see variable "show_basket"
                # of the  active tab
                if tab['href'].startswith(request_path):
                    active_tab = tab.copy()

    # calculate minimum number of tabs
    min_num_tabs = len(context['fixed_tabs'])

    if not context.get('connect_fixed_tabs_with_extra_tabs', True):  # default is to connect
        min_num_tabs += 1  # avoid double bracket between fixed and extra tabs -> no connection

    local_context = dict(tabs=tabs, min_num_tabs=min_num_tabs, active_tab=active_tab)
    if 'exception' in context:
        # we want to show exceptions in an own tab
        local_context['error'] = 'error'  # must be not empty for if in template

    return local_context



import importlib.metadata
import logging

from topobank.plugins import PluginConfig

try:
    __version__ = importlib.metadata.version('ce-ui')
except importlib.metadata.PackageNotFoundError:
    __version__ = '0.0.0'

_log = logging.Logger(__file__)


class CEUIPluginConfig(PluginConfig):
    name = 'ce_ui'
    verbose_name = "contact.engineering"

    class TopobankPluginMeta:
        name = "contact.engineering"
        version = __version__
        description = """
        The contact.engineering user interface
        """
        logo = "ce_ui/static/images/ce_logo.svg"
        restricted = False  # Accessible for all users, without permissions

    def ready(self):
        # make sure the functions are registered now

        # noinspection PyUnresolvedReferences
        import ce_ui.signals  # noqa: F401
